
import string
import random
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.experiments.mutations.types.base.mutation import Mutation
from hedra.core.experiments.mutations.types.base.mutation_type import MutationType
from typing import Union, Optional, List, Tuple
from .validator import DeformHeaderValidator


Request = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action
]


class DeformHeader(Mutation):

    def __init__(
        self, 
        name: str, 
        chance: float,
        *targets: Tuple[str, ...],
        header_name: str=None,
        header_value: Optional[str]=None,
        deformation_length: int=5,
        character_pool: Optional[List[str]]=[]
    ) -> None:
        super().__init__(
            name, 
            chance,
            MutationType.DEFORM_HEADER,
            *targets
        )

        validated_mutation = DeformHeaderValidator(
            header_name=header_name,
            deformation_length=deformation_length,
            character_pool=character_pool,
            header_value=header_value
        )

        self.header_name = validated_mutation.header_name
        self.header_value = validated_mutation.header_value
        self.deformation_length = validated_mutation.deformation_length

        if validated_mutation.character_pool:
            self.character_pool = ''.join(validated_mutation.character_pool)
            
        else:
            self.character_pool = ''.join([
                string.ascii_letters,
                string.digits,
                string.hexdigits,
                string.octdigits,
                string.punctuation,
                string.whitespace
            ])

        self.header_mutation = ''.join(
            random.choices(
                self.character_pool, 
                k=self.deformation_length
            )
        )

    async def mutate(self, action: Request=None) -> Request:

        chance_roll = random.uniform(0, 1)
        if chance_roll <= self.chance:
            return action

        
        mutated_header_name = f'{self.header_name} {self.header_mutation}'

        header_value = action.headers.get(self.header_name)

        if self.header_value:
            header_value = self.header_value

        del action._headers[self.header_name]

        action._headers[mutated_header_name] = header_value
        action._header_items = list(action._headers.items())

        action._setup_headers()
        
        return action
    
    def copy(self):
        return DeformHeader(
            self.name,
            self.chance,
            *list(self.targets),
            header_name=self.header_name,
            header_value=self.header_name,
            deformation_length=self.deformation_length,
            character_pool=self.character_pool
        )


from typing import Optional
from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool
)


class RedisConnectorConfig(BaseModel):
    host: StrictStr='localhost:6379'
    username: Optional[StrictStr]
    password: Optional[StrictStr]
    database: StrictInt=0
    channel: StrictStr
    channel_type: StrictStr='pipeline'
    secure: StrictBool=False
    connector_type: ConnectorType=ConnectorType.Redis


from __future__ import annotations
import asyncio
import re
from enum import Enum, IntEnum
from collections import deque
from typing import (
    Deque, 
    Dict, 
    List, 
    Optional, 
    cast, 
    Set, 
    FrozenSet, 
    Callable,
    Tuple,
    Any,
    Union,
    Text,
)


def dummy_unit_encode(val: int):
    return b''


def dummy_stream_is_unidirectional(val: int):
    return False
    

try:

    import pylsqpack
    from aioquic.quic.connection import QuicConnection, NetworkAddress
    from aioquic.quic.events import (
        ConnectionTerminated,
        ConnectionIdRetired,
        ConnectionIdIssued,
        HandshakeCompleted,
        PingAcknowledged
    )
    from aioquic.h3.events import (
        DatagramReceived,
        DataReceived,
        H3Event,
        Headers,
        HeadersReceived,
        PushPromiseReceived,
        WebTransportStreamDataReceived,
    )
    from aioquic.buffer import (
        UINT_VAR_MAX_SIZE,
        Buffer, 
        BufferReadError, 
        encode_uint_var
    )

    from aioquic.quic.events import (
        DatagramFrameReceived, 
        QuicEvent, 
        StreamDataReceived
    )

    from aioquic.quic.connection import stream_is_unidirectional

except ImportError:
    pylsqpack = object
    aioquic = object
    QuicConnection = object
    ConnectionTerminated = object
    ConnectionIdRetired = object
    ConnectionIdIssued = object
    HandshakeCompleted = object
    PingAcknowledged = object
    DatagramReceived = object
    DataReceived = object
    H3Event = object
    Headers = Any
    HeadersReceived = object
    PushPromiseReceived = object
    WebTransportStreamDataReceived = object
    NetworkAddress = Any
    UINT_VAR_MAX_SIZE = 0 
    Buffer = object
    BufferReadError = object 
    encode_uint_var = dummy_unit_encode
    DatagramFrameReceived = object
    QuicEvent = object
    StreamDataReceived = object
    stream_is_unidirectional = dummy_stream_is_unidirectional


QuicConnectionIdHandler = Callable[[bytes], None]
QuicStreamHandler = Callable[[asyncio.StreamReader, asyncio.StreamWriter], None]


USER_AGENT = "hedra/client"
RESERVED_SETTINGS = (0x0, 0x2, 0x3, 0x4, 0x5)
UPPERCASE = re.compile(b"[A-Z]")


class ErrorCode(IntEnum):
    H3_NO_ERROR = 0x100
    H3_GENERAL_PROTOCOL_ERROR = 0x101
    H3_INTERNAL_ERROR = 0x102
    H3_STREAM_CREATION_ERROR = 0x103
    H3_CLOSED_CRITICAL_STREAM = 0x104
    H3_FRAME_UNEXPECTED = 0x105
    H3_FRAME_ERROR = 0x106
    H3_EXCESSIVE_LOAD = 0x107
    H3_ID_ERROR = 0x108
    H3_SETTINGS_ERROR = 0x109
    H3_MISSING_SETTINGS = 0x10A
    H3_REQUEST_REJECTED = 0x10B
    H3_REQUEST_CANCELLED = 0x10C
    H3_REQUEST_INCOMPLETE = 0x10D
    H3_MESSAGE_ERROR = 0x10E
    H3_CONNECT_ERROR = 0x10F
    H3_VERSION_FALLBACK = 0x110
    QPACK_DECOMPRESSION_FAILED = 0x200
    QPACK_ENCODER_STREAM_ERROR = 0x201
    QPACK_DECODER_STREAM_ERROR = 0x202


class Setting(IntEnum):
    QPACK_MAX_TABLE_CAPACITY = 0x1
    MAX_FIELD_SECTION_SIZE = 0x6
    QPACK_BLOCKED_STREAMS = 0x7

    # https://datatracker.ietf.org/doc/html/rfc9220#section-5
    ENABLE_CONNECT_PROTOCOL = 0x8
    # https://datatracker.ietf.org/doc/html/draft-ietf-masque-h3-datagram-05#section-9.1
    H3_DATAGRAM = 0xFFD277
    # https://datatracker.ietf.org/doc/html/draft-ietf-webtrans-http2-02#section-10.1
    ENABLE_WEBTRANSPORT = 0x2B603742

    # Dummy setting to check it is correctly ignored by the peer.
    # https://datatracker.ietf.org/doc/html/rfc9114#section-7.2.4.1
    DUMMY = 0x21


def encode_frame(frame_type: int, frame_data: bytes) -> bytes:
    frame_length = len(frame_data)
    buf = Buffer(capacity=frame_length + 2 * UINT_VAR_MAX_SIZE)
    buf.push_uint_var(frame_type)
    buf.push_uint_var(frame_length)
    buf.push_bytes(frame_data)
    return buf.data


def validate_headers(
    headers: Headers,
    allowed_pseudo_headers: FrozenSet[bytes],
    required_pseudo_headers: FrozenSet[bytes],
) -> None:
    after_pseudo_headers = False
    authority: Optional[bytes] = None
    path: Optional[bytes] = None
    scheme: Optional[bytes] = None
    seen_pseudo_headers: Set[bytes] = set()
    for key, value in headers:
        if UPPERCASE.search(key):
            raise MessageError("Header %r contains uppercase letters" % key)

        if key.startswith(b":"):
            # pseudo-headers
            if after_pseudo_headers:
                raise MessageError(
                    "Pseudo-header %r is not allowed after regular headers" % key
                )
            if key not in allowed_pseudo_headers:
                raise MessageError("Pseudo-header %r is not valid" % key)
            if key in seen_pseudo_headers:
                raise MessageError("Pseudo-header %r is included twice" % key)
            seen_pseudo_headers.add(key)

            # store value
            if key == b":authority":
                authority = value
            elif key == b":path":
                path = value
            elif key == b":scheme":
                scheme = value
        else:
            # regular headers
            after_pseudo_headers = True

    # check required pseudo-headers are present
    missing = required_pseudo_headers.difference(seen_pseudo_headers)
    if missing:
        raise MessageError("Pseudo-headers %s are missing" % sorted(missing))

    if scheme in (b"http", b"https"):
        if not authority:
            raise MessageError("Pseudo-header b':authority' cannot be empty")
        if not path:
            raise MessageError("Pseudo-header b':path' cannot be empty")


class ProtocolError(Exception):
    """
    Base class for protocol errors.

    These errors are not exposed to the API user, they are handled
    in :meth:`H3Connection.handle_event`.
    """

    error_code = ErrorCode.H3_GENERAL_PROTOCOL_ERROR

    def __init__(self, reason_phrase: str = ""):
        self.reason_phrase = reason_phrase

class QpackDecompressionFailed(ProtocolError):
    error_code = ErrorCode.QPACK_DECOMPRESSION_FAILED


class QpackDecoderStreamError(ProtocolError):
    error_code = ErrorCode.QPACK_DECODER_STREAM_ERROR


class QpackEncoderStreamError(ProtocolError):
    error_code = ErrorCode.QPACK_ENCODER_STREAM_ERROR


class ClosedCriticalStream(ProtocolError):
    error_code = ErrorCode.H3_CLOSED_CRITICAL_STREAM


class FrameUnexpected(ProtocolError):
    error_code = ErrorCode.H3_FRAME_UNEXPECTED


class MessageError(ProtocolError):
    error_code = ErrorCode.H3_MESSAGE_ERROR


class MissingSettingsError(ProtocolError):
    error_code = ErrorCode.H3_MISSING_SETTINGS


class SettingsError(ProtocolError):
    error_code = ErrorCode.H3_SETTINGS_ERROR


class StreamCreationError(ProtocolError):
    error_code = ErrorCode.H3_STREAM_CREATION_ERROR

class FrameType(IntEnum):
    DATA = 0x0
    HEADERS = 0x1
    PRIORITY = 0x2
    CANCEL_PUSH = 0x3
    SETTINGS = 0x4
    PUSH_PROMISE = 0x5
    GOAWAY = 0x7
    MAX_PUSH_ID = 0xD
    DUPLICATE_PUSH = 0xE
    WEBTRANSPORT_STREAM = 0x41

class StreamType(IntEnum):
    CONTROL = 0
    PUSH = 1
    QPACK_ENCODER = 2
    QPACK_DECODER = 3
    WEBTRANSPORT = 0x54

class HeadersState(Enum):
    INITIAL = 0
    AFTER_HEADERS = 1
    AFTER_TRAILERS = 2

class H3Stream:

    __slots__ =(
        'blocked',
        'blocked_frame_size',
        'buffer',
        'ended',
        'frame_size',
        'frame_type',
        'headers_recv_state',
        'headers_send_state',
        'push_id',
        'session_id',
        'stream_id',
        'stream_type'
    )

    def __init__(self, stream_id: int) -> None:
        self.blocked = False
        self.blocked_frame_size: Optional[int] = None
        self.buffer = b""
        self.ended = False
        self.frame_size: Optional[int] = None
        self.frame_type: Optional[int] = None
        self.headers_recv_state: HeadersState = HeadersState.INITIAL
        self.headers_send_state: HeadersState = HeadersState.INITIAL
        self.push_id: Optional[int] = None
        self.session_id: Optional[int] = None
        self.stream_id = stream_id
        self.stream_type: Optional[int] = None


class ResponseFrameCollection:

    __slots__ = (
        'headers_frame',
        'body'
    )

    def __init__(self) -> None:
        self.headers_frame: HeadersReceived = None
        self.body = bytearray()


class QuicStreamAdapter(asyncio.Transport):

    __slots__ = (
        'protocol',
        'stream_id'
    )

    def __init__(self, protocol: QuicProtocol, stream_id: int):
        self.protocol = protocol
        self.stream_id = stream_id

    def can_write_eof(self) -> bool:
        return True

    def get_extra_info(self, name: str, default: Any = None) -> Any:
        """
        Get information about the underlying QUIC stream.
        """
        if name == "stream_id":
            return self.stream_id

    def write(self, data):
        self.protocol._quic.send_stream_data(self.stream_id, data)
        self.protocol._transmit_soon()

    def write_eof(self):
        self.protocol._quic.send_stream_data(self.stream_id, b"", end_stream=True)
        self.protocol._transmit_soon()


class QuicProtocol(asyncio.DatagramProtocol):

    __slots__ = (
        '_loop',
        '_request_waiter',
        '_closed',
        '_connected',
        '_connected_waiter',
        '_ping_waiters',
        '_quic',
        '_stream_readers',
        '_timer',
        '_timer_at',
        '_transmit_task',
        '_transport',
        '_connection_id_issued_handler',
        '_connection_id_retired_handler',
        '_connection_terminated_handler',
        '_stream_handler',
        'pushes',
        '_request_events',
        '_request_waiter',
        '_stream',
        '_is_done',
        '_max_table_capacity',
        '_blocked_streams',
        '_decoder',
        '_decoder_bytes_received',
        '_decoder_bytes_sent',
        '_encoder',
        '_encoder_bytes_received',
        '_encoder_bytes_sent',
        '_settings_received',
        '_stream',
        '_max_push_id',
        '_next_push_id',
        '_local_control_stream_id',
        '_local_decoder_stream_id',
        '_local_encoder_stream_id',
        '_peer_control_stream_id',
        '_peer_decoder_stream_id',
        '_peer_encoder_stream_id',
        '_received_settings',
        '_sent_settings',
        'responses',
        '_is_closed'
    )

    def __init__(
        self, quic: QuicConnection, 
        stream_handler: Optional[QuicStreamHandler] = None,
        loop: asyncio.AbstractEventLoop = None
    ):
        
        self._loop = loop

        self._request_waiter: Dict[int, asyncio.Future[Deque[H3Event]]] = {}
        self._closed = asyncio.Event()
        self._connected = False
        self._connected_waiter: Optional[asyncio.Future[None]] = None
        self._ping_waiters: Dict[int, asyncio.Future[None]] = {}
        self._quic = quic
        self._stream_readers: Dict[int, asyncio.StreamReader] = {}
        self._timer: Optional[asyncio.TimerHandle] = None
        self._timer_at: Optional[float] = None
        self._transmit_task: Optional[asyncio.Handle] = None
        self._transport: Optional[asyncio.DatagramTransport] = None

        # callbacks
        self._connection_id_issued_handler: QuicConnectionIdHandler = lambda c: None
        self._connection_id_retired_handler: QuicConnectionIdHandler = lambda c: None
        self._connection_terminated_handler: Callable[[], None] = lambda: None
        if stream_handler is not None:
            self._stream_handler = stream_handler
        else:
            self._stream_handler = lambda r, w: None

        self.pushes: Dict[int, Deque[H3Event]] = {}
        self._request_events: Dict[int, Deque[H3Event]] = {}
        self._request_waiter: Dict[int, asyncio.Future[Deque[H3Event]]] = {}
        self._stream: Dict[int, H3Stream] = {}
        self._is_done = False
        self._max_table_capacity = 4096
        self._blocked_streams = 128

        self._decoder = pylsqpack.Decoder(
            self._max_table_capacity, self._blocked_streams
        )
        self._decoder_bytes_received = 0
        self._decoder_bytes_sent = 0
        self._encoder = pylsqpack.Encoder()
        self._encoder_bytes_received = 0
        self._encoder_bytes_sent = 0
        self._settings_received = False
        self._stream: Dict[int, H3Stream] = {}

        self._max_push_id: Optional[int] = None
        self._next_push_id: int = 0

        self._local_control_stream_id: Optional[int] = None
        self._local_decoder_stream_id: Optional[int] = None
        self._local_encoder_stream_id: Optional[int] = None

        self._peer_control_stream_id: Optional[int] = None
        self._peer_decoder_stream_id: Optional[int] = None
        self._peer_encoder_stream_id: Optional[int] = None
        self._received_settings: Optional[Dict[int, int]] = None
        self._sent_settings: Optional[Dict[int, int]] = None
        self.responses: Dict[int, ResponseFrameCollection] = {} 
        self._is_closed = False
        

    def init_connection(self) -> None:
        # send our settings
        self._local_control_stream_id = self._create_uni_stream(StreamType.CONTROL)
        self._sent_settings = {
            Setting.QPACK_MAX_TABLE_CAPACITY: self._max_table_capacity,
            Setting.QPACK_BLOCKED_STREAMS: self._blocked_streams,
            Setting.ENABLE_CONNECT_PROTOCOL: 1,
            Setting.DUMMY: 1,
        }

        buf = Buffer(capacity=1024)

        for setting, value in self._sent_settings.items():
            buf.push_uint_var(setting)
            buf.push_uint_var(value)

        self._quic.send_stream_data(
            self._local_control_stream_id,
            encode_frame(FrameType.SETTINGS, buf.data),
        )
        if self._max_push_id is not None:
            self._quic.send_stream_data(
                self._local_control_stream_id,
                encode_frame(FrameType.MAX_PUSH_ID, encode_uint_var(self._max_push_id)),
            )

        # create encoder and decoder streams
        self._local_encoder_stream_id = self._create_uni_stream(
            StreamType.QPACK_ENCODER
        )
        self._local_decoder_stream_id = self._create_uni_stream(
            StreamType.QPACK_DECODER
        )

    def _create_uni_stream(
        self, stream_type: int, push_id: Optional[int] = None
    ) -> int:
        """
        Create an unidirectional stream of the given type.
        """
        stream_id = self._quic.get_next_available_stream_id(is_unidirectional=True)
        self._quic.send_stream_data(stream_id, encode_uint_var(stream_type))
        return stream_id

    def change_connection_id(self) -> None:
        """
        Change the connection ID used to communicate with the peer.

        The previous connection ID will be retired.
        """
        self._quic.change_connection_id()
        self.transmit()

    def close(self) -> None:
        """
        Close the connection.
        """
        self._is_closed = True
        self._quic.close()
        self.transmit()

    def connect(self, addr: NetworkAddress) -> None:
        """
        Initiate the TLS handshake.

        This method can only be called for clients and a single time.
        """
        self._quic.connect(addr, now=self._loop.time())
        self.transmit()

    
    def _get_or_create_stream(self, stream_id: int) -> H3Stream:
        if stream_id not in self._stream:
            self._stream[stream_id] = H3Stream(stream_id)
        return self._stream[stream_id]
    

    async def create_stream(
        self, is_unidirectional: bool = False
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """
        Create a QUIC stream and return a pair of (reader, writer) objects.

        The returned reader and writer objects are instances of :class:`asyncio.StreamReader`
        and :class:`asyncio.StreamWriter` classes.
        """
        stream_id = self._quic.get_next_available_stream_id(
            is_unidirectional=is_unidirectional
        )
        return self._create_stream(stream_id)

    def request_key_update(self) -> None:
        """
        Request an update of the encryption keys.
        """
        self._quic.request_key_update()
        self.transmit()

    async def ping(self) -> None:
        """
        Ping the peer and wait for the response.
        """
        waiter = self._loop.create_future()
        uid = id(waiter)
        self._ping_waiters[uid] = waiter
        self._quic.send_ping(uid)
        self.transmit()
        await asyncio.shield(waiter)

    def transmit(self) -> None:
        """
        Send pending datagrams to the peer and arm the timer if needed.
        """
        self._transmit_task = None

        # send datagrams
        for data, addr in self._quic.datagrams_to_send(now=self._loop.time()):
            self._transport.sendto(data, addr)

        # re-arm timer
        timer_at = self._quic.get_timer()
        if self._timer is not None and self._timer_at != timer_at:
            self._timer.cancel()
            self._timer = None
        if self._timer is None and timer_at is not None:
            self._timer = self._loop.call_at(timer_at, self._handle_timer)
        self._timer_at = timer_at

    async def wait_closed(self) -> None:
        """
        Wait for the connection to be closed.
        """
        await self._closed.wait()

    async def wait_connected(self) -> None:
        """
        Wait for the TLS handshake to complete.
        """
        assert self._connected_waiter is None, "already awaiting connected"
        if not self._connected:
            self._connected_waiter = self._loop.create_future()
            await self._connected_waiter

    # asyncio.Transport

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self._transport = cast(asyncio.DatagramTransport, transport)

    def datagram_received(self, data: Union[bytes, Text], addr: NetworkAddress) -> None:
        self._quic.receive_datagram(cast(bytes, data), addr, now=self._loop.time())
        self._process_events()
        self.transmit()

    # overridable
    def quic_event_received(self, event: QuicEvent) -> None:
        #  pass event to the HTTP layer'
        events = []
        if not self._is_done:
            try:
                if isinstance(event, StreamDataReceived):
                    stream_id = event.stream_id
                    stream = self._get_or_create_stream(stream_id)
                    if stream_is_unidirectional(stream_id):

                        http_events: List[H3Event] = []

                        stream.buffer += event.data
                        if event.end_stream:
                            stream.ended = True

                        buf = Buffer(data=stream.buffer)
                        consumed = 0
                        unblocked_streams: Set[int] = set()

                        while (
                            stream.stream_type
                            in (StreamType.PUSH, StreamType.CONTROL, StreamType.WEBTRANSPORT)
                            or not buf.eof()
                        ):
                            # fetch stream type for unidirectional streams
                            if stream.stream_type is None:
                                try:
                                    stream.stream_type = buf.pull_uint_var()
                                except BufferReadError:
                                    break
                                consumed = buf.tell()

                                # check unicity
                                if stream.stream_type == StreamType.CONTROL:
                                    if self._peer_control_stream_id is not None:
                                        raise StreamCreationError("Only one control stream is allowed")
                                    self._peer_control_stream_id = stream.stream_id
                                elif stream.stream_type == StreamType.QPACK_DECODER:
                                    if self._peer_decoder_stream_id is not None:
                                        raise StreamCreationError(
                                            "Only one QPACK decoder stream is allowed"
                                        )
                                    self._peer_decoder_stream_id = stream.stream_id
                                elif stream.stream_type == StreamType.QPACK_ENCODER:
                                    if self._peer_encoder_stream_id is not None:
                                        raise StreamCreationError(
                                            "Only one QPACK encoder stream is allowed"
                                        )
                                    self._peer_encoder_stream_id = stream.stream_id



                            if stream.stream_type == StreamType.CONTROL:
                                if event.end_stream:
                                    raise ClosedCriticalStream("Closing control stream is not allowed")

                                # fetch next frame
                                try:
                                    frame_type = buf.pull_uint_var()
                                    frame_length = buf.pull_uint_var()
                                    frame_data = buf.pull_bytes(frame_length)
                                except BufferReadError:
                                    break
                                consumed = buf.tell()

                                if frame_type != FrameType.SETTINGS and not self._settings_received:
                                    raise MissingSettingsError

                                if frame_type == FrameType.SETTINGS:
                                    if self._settings_received:
                                        raise FrameUnexpected("SETTINGS have already been received")
                            
                                    buf = Buffer(data=frame_data)
                                    settings: Dict[int, int] = {}
                                    while not buf.eof():
                                        setting = buf.pull_uint_var()
                                        value = buf.pull_uint_var()
                                        if setting in RESERVED_SETTINGS:
                                            raise SettingsError("Setting identifier 0x%x is reserved" % setting)
                                        if setting in settings:
                                            raise SettingsError("Setting identifier 0x%x is included twice" % setting)
                                        settings[setting] = value   

                                    for setting in [
                                        Setting.ENABLE_CONNECT_PROTOCOL,
                                        Setting.ENABLE_WEBTRANSPORT,
                                        Setting.H3_DATAGRAM,
                                    ]:
                                        if setting in settings and settings[setting] not in (0, 1):
                                            raise SettingsError(f"{setting.name} setting must be 0 or 1")

                                    if (
                                        settings.get(Setting.H3_DATAGRAM) == 1
                                        and self._quic._remote_max_datagram_frame_size is None
                                    ):
                                        raise SettingsError(
                                            "H3_DATAGRAM requires max_datagram_frame_size transport parameter"
                                        )

                                    if (
                                        settings.get(Setting.ENABLE_WEBTRANSPORT) == 1
                                        and settings.get(Setting.H3_DATAGRAM) != 1
                                    ):
                                        raise SettingsError("ENABLE_WEBTRANSPORT requires H3_DATAGRAM")

                                    self._received_settings = settings
                                    encoder = self._encoder.apply_settings(
                                        max_table_capacity=settings.get(Setting.QPACK_MAX_TABLE_CAPACITY, 0),
                                        blocked_streams=settings.get(Setting.QPACK_BLOCKED_STREAMS, 0),
                                    )
                                    self._quic.send_stream_data(self._local_encoder_stream_id, encoder)
                                    self._settings_received = True

                                elif frame_type in (
                                    FrameType.DATA,
                                    FrameType.HEADERS,
                                    FrameType.PUSH_PROMISE,
                                    FrameType.DUPLICATE_PUSH,
                                ):
                                    raise FrameUnexpected("Invalid frame type on control stream")

                            elif stream.stream_type == StreamType.PUSH:
                                # fetch push id
                                if stream.push_id is None:
                                    try:
                                        stream.push_id = buf.pull_uint_var()
                                    except BufferReadError:
                                        break
                                    consumed = buf.tell()

                                # remove processed data from buffer
                                stream.buffer = stream.buffer[consumed:]

                                return self._receive_request_or_push_data(stream, b"", event.end_stream)
                            elif stream.stream_type == StreamType.WEBTRANSPORT:
                                # fetch session id
                                if stream.session_id is None:
                                    try:
                                        stream.session_id = buf.pull_uint_var()
                                    except BufferReadError:
                                        break
                                    consumed = buf.tell()

                                frame_data = stream.buffer[consumed:]
                                stream.buffer = b""

                                if frame_data or event.end_stream:
                                    http_events.append(
                                        WebTransportStreamDataReceived(
                                            data=frame_data,
                                            session_id=stream.session_id,
                                            stream_ended=stream.ended,
                                            stream_id=stream.stream_id,
                                        )
                                    )
                                return http_events
                            elif stream.stream_type == StreamType.QPACK_DECODER:
                                # feed unframed data to decoder
                                data = buf.pull_bytes(buf.capacity - buf.tell())
                                consumed = buf.tell()
                                try:
                                    self._encoder.feed_decoder(data)
                                except pylsqpack.DecoderStreamError as exc:
                                    raise QpackDecoderStreamError() from exc
                                self._decoder_bytes_received += len(data)
                            elif stream.stream_type == StreamType.QPACK_ENCODER:
                                # feed unframed data to encoder
                                data = buf.pull_bytes(buf.capacity - buf.tell())
                                consumed = buf.tell()
                                try:
                                    unblocked_streams.update(self._decoder.feed_encoder(data))
                                except pylsqpack.EncoderStreamError as exc:
                                    raise QpackEncoderStreamError() from exc
                                self._encoder_bytes_received += len(data)
                            else:
                                # unknown stream type, discard data
                                buf.seek(buf.capacity)
                                consumed = buf.tell()

                        # remove processed data from buffer
                        stream.buffer = stream.buffer[consumed:]

                        # process unblocked streams
                        for stream_id in unblocked_streams:
                            stream = self._stream[stream_id]

                            # resume headers
                            http_events.extend(
                                self._handle_request_or_push_frame(
                                    frame_type=FrameType.HEADERS,
                                    frame_data=None,
                                    stream=stream,
                                    stream_ended=stream.ended and not stream.buffer,
                                )
                            )
                            stream.blocked = False
                            stream.blocked_frame_size = None

                            # resume processing
                            if stream.buffer:
                                http_events.extend(
                                    self._receive_request_or_push_data(stream, b"", stream.ended)
                                )

                        events = http_events

                    else:
                        events = self._receive_request_or_push_data(
                            stream, event.data, event.end_stream
                        )
                elif isinstance(event, DatagramFrameReceived):
                    buf = Buffer(data=event.data)
                    try:
                        flow_id = buf.pull_uint_var()
                    except BufferReadError:
                        raise ProtocolError("Could not parse flow ID")
                    
                    events = [DatagramReceived(
                        data=event.data[buf.tell() :], 
                        flow_id=flow_id
                    )]

            except ProtocolError as exc:
                self._is_done = True
                self._quic.close(
                    error_code=exc.error_code, reason_phrase=exc.reason_phrase
                )

        for http_event in events:

            if isinstance(http_event, HeadersReceived):

                if self.responses.get(http_event.stream_id) is None:
                    self.responses[http_event.stream_id] = ResponseFrameCollection()

                self.responses[http_event.stream_id].headers_frame = http_event

                if http_event.push_id in self.pushes:
                    # push
                    self.pushes[http_event.push_id].append(http_event)


            elif isinstance(http_event, DataReceived):

                if self.responses.get(http_event.stream_id) is None:
                    self.responses[http_event.stream_id] = ResponseFrameCollection()

                self.responses[http_event.stream_id].body.extend(http_event.data)
                request_waiter = self._request_waiter.get(http_event.stream_id)

                if http_event.stream_ended and request_waiter is None:
                    raise Exception('Err. - Stream failed')
                
                elif http_event.stream_ended and request_waiter.done() is False:
                    request_waiter.set_result(self.responses.pop(http_event.stream_id))
                
                if http_event.push_id in self.pushes:
                    # push
                    self.pushes[http_event.push_id].append(http_event)

            elif isinstance(event, PushPromiseReceived):
                self.pushes[event.push_id] = deque()
                self.pushes[event.push_id].append(event)
    
    def _receive_request_or_push_data(
        self, stream: H3Stream, data: bytes, stream_ended: bool
    ) -> List[H3Event]:
        """
        Handle data received on a request or push stream.
        """
        http_events: List[H3Event] = []

        stream.buffer += data
        if stream_ended:
            stream.ended = True
        if stream.blocked:
            return http_events

        # shortcut for WEBTRANSPORT_STREAM frame fragments
        if (
            stream.frame_type == FrameType.WEBTRANSPORT_STREAM
            and stream.session_id is not None
        ):
            http_events.append(
                WebTransportStreamDataReceived(
                    data=stream.buffer,
                    session_id=stream.session_id,
                    stream_id=stream.stream_id,
                    stream_ended=stream_ended,
                )
            )
            stream.buffer = b""
            return http_events

        # shortcut for DATA frame fragments
        if (
            stream.frame_type == FrameType.DATA
            and stream.frame_size is not None
            and len(stream.buffer) < stream.frame_size
        ):
            http_events.append(
                DataReceived(
                    data=stream.buffer,
                    push_id=stream.push_id,
                    stream_id=stream.stream_id,
                    stream_ended=False,
                )
            )
            stream.frame_size -= len(stream.buffer)
            stream.buffer = b""
            return http_events

        # handle lone FIN
        if stream_ended and not stream.buffer:
            http_events.append(
                DataReceived(
                    data=b"",
                    push_id=stream.push_id,
                    stream_id=stream.stream_id,
                    stream_ended=True,
                )
            )
            return http_events

        buf = Buffer(data=stream.buffer)
        consumed = 0

        while not buf.eof():
            # fetch next frame header
            if stream.frame_size is None:
                try:
                    stream.frame_type = buf.pull_uint_var()
                    stream.frame_size = buf.pull_uint_var()
                except BufferReadError:
                    break
                consumed = buf.tell()

                # WEBTRANSPORT_STREAM frames last until the end of the stream
                if stream.frame_type == FrameType.WEBTRANSPORT_STREAM:
                    stream.session_id = stream.frame_size
                    stream.frame_size = None

                    frame_data = stream.buffer[consumed:]
                    stream.buffer = b""

                    if frame_data or stream_ended:
                        http_events.append(
                            WebTransportStreamDataReceived(
                                data=frame_data,
                                session_id=stream.session_id,
                                stream_id=stream.stream_id,
                                stream_ended=stream_ended,
                            )
                        )
                    return http_events

            # check how much data is available
            chunk_size = min(stream.frame_size, buf.capacity - consumed)
            if stream.frame_type != FrameType.DATA and chunk_size < stream.frame_size:
                break

            # read available data
            frame_data = buf.pull_bytes(chunk_size)
            frame_type = stream.frame_type
            consumed = buf.tell()

            # detect end of frame
            stream.frame_size -= chunk_size
            if not stream.frame_size:
                stream.frame_size = None
                stream.frame_type = None

            try:
                http_events.extend(
                    self._handle_request_or_push_frame(
                        frame_type=frame_type,
                        frame_data=frame_data,
                        stream=stream,
                        stream_ended=stream.ended and buf.eof(),
                    )
                )
            except pylsqpack.StreamBlocked:
                stream.blocked = True
                stream.blocked_frame_size = len(frame_data)
                break

        # remove processed data from buffer
        stream.buffer = stream.buffer[consumed:]

        return http_events

    def _create_stream(
        self, stream_id: int
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        adapter = QuicStreamAdapter(self, stream_id)
        reader = asyncio.StreamReader()
        writer = asyncio.StreamWriter(adapter, None, reader, self._loop)
        self._stream_readers[stream_id] = reader
        return reader, writer

    def _handle_timer(self) -> None:
        now = max(self._timer_at, self._loop.time())
        self._timer = None
        self._timer_at = None
        self._quic.handle_timer(now=now)
        self._process_events()
        self.transmit()

    def _process_events(self) -> None:
        event = self._quic.next_event()
        while event is not None:
            if isinstance(event, ConnectionIdIssued):
                self._connection_id_issued_handler(event.connection_id)
            elif isinstance(event, ConnectionIdRetired):
                self._connection_id_retired_handler(event.connection_id)
            elif isinstance(event, ConnectionTerminated):
                self._connection_terminated_handler()

                # abort connection waiter
                if self._connected_waiter is not None:
                    waiter = self._connected_waiter
                    self._connected_waiter = None
                    waiter.set_exception(ConnectionError)

                # abort ping waiters
                for waiter in self._ping_waiters.values():
                    waiter.set_exception(ConnectionError)
                self._ping_waiters.clear()

                self._closed.set()
            elif isinstance(event, HandshakeCompleted):
                if self._connected_waiter is not None:
                    waiter = self._connected_waiter
                    self._connected = True
                    self._connected_waiter = None
                    waiter.set_result(None)
            elif isinstance(event, PingAcknowledged):
                waiter = self._ping_waiters.pop(event.uid, None)
                if waiter is not None:
                    waiter.set_result(None)
            self.quic_event_received(event)
            event = self._quic.next_event()

    def _transmit_soon(self) -> None:
        if self._transmit_task is None:
            self._transmit_task = self._loop.call_soon(self.transmit)

    def _handle_request_or_push_frame(
        self,
        frame_type: int,
        frame_data: Optional[bytes],
        stream: H3Stream,
        stream_ended: bool,
    ) -> List[H3Event]:
        """
        Handle a frame received on a request or push stream.
        """
        http_events: List[H3Event] = []

        if frame_type == FrameType.DATA:
            # check DATA frame is allowed
            if stream.headers_recv_state != HeadersState.AFTER_HEADERS:
                raise FrameUnexpected("DATA frame is not allowed in this state")

            if stream_ended or frame_data:
                http_events.append(
                    DataReceived(
                        data=frame_data,
                        push_id=stream.push_id,
                        stream_ended=stream_ended,
                        stream_id=stream.stream_id,
                    )
                )
        elif frame_type == FrameType.HEADERS:
            # check HEADERS frame is allowed
            if stream.headers_recv_state == HeadersState.AFTER_TRAILERS:
                raise FrameUnexpected("HEADERS frame is not allowed in this state")

            # try to decode HEADERS, may raise pylsqpack.StreamBlocked
            headers = self._decode_headers(stream.stream_id, frame_data)

            # validate headers
            if stream.headers_recv_state == HeadersState.INITIAL:
                validate_headers(
                    headers,
                    allowed_pseudo_headers=frozenset((b":status",)),
                    required_pseudo_headers=frozenset((b":status",)),
                )

            else:
                validate_headers(
                    headers,
                    allowed_pseudo_headers=frozenset(),
                    required_pseudo_headers=frozenset(),
                )


            # update state and emit headers
            if stream.headers_recv_state == HeadersState.INITIAL:
                stream.headers_recv_state = HeadersState.AFTER_HEADERS
            else:
                stream.headers_recv_state = HeadersState.AFTER_TRAILERS
            http_events.append(
                HeadersReceived(
                    headers=headers,
                    push_id=stream.push_id,
                    stream_id=stream.stream_id,
                    stream_ended=stream_ended,
                )
            )
        elif frame_type == FrameType.PUSH_PROMISE and stream.push_id is None:
            
            frame_buf = Buffer(data=frame_data)
            push_id = frame_buf.pull_uint_var()
            headers = self._decode_headers(
                stream.stream_id, frame_data[frame_buf.tell() :]
            )

            # validate headers
            validate_headers(
                headers,
                allowed_pseudo_headers=frozenset(
                    (b":method", b":scheme", b":authority", b":path")
                ),
                required_pseudo_headers=frozenset(
                    (b":method", b":scheme", b":authority", b":path")
                ),
            )

            # emit event
            http_events.append(
                PushPromiseReceived(
                    headers=headers, push_id=push_id, stream_id=stream.stream_id
                )
            )
        elif frame_type in (
            FrameType.PRIORITY,
            FrameType.CANCEL_PUSH,
            FrameType.SETTINGS,
            FrameType.PUSH_PROMISE,
            FrameType.GOAWAY,
            FrameType.MAX_PUSH_ID,
            FrameType.DUPLICATE_PUSH,
        ):
            raise FrameUnexpected(
                "Invalid frame type on request stream"
                if stream.push_id is None
                else "Invalid frame type on push stream"
            )

        return http_events
    
    def _decode_headers(self, stream_id: int, frame_data: Optional[bytes]) -> Headers:
        """
        Decode a HEADERS block and send decoder updates on the decoder stream.

        This is called with frame_data=None when a stream becomes unblocked.
        """
        try:
            if frame_data is None:
                decoder, headers = self._decoder.resume_header(stream_id)
            else:
                decoder, headers = self._decoder.feed_header(stream_id, frame_data)
            self._decoder_bytes_sent += len(decoder)
            self._quic.send_stream_data(self._local_decoder_stream_id, decoder)
        except pylsqpack.DecompressionFailed as exc:
            raise QpackDecompressionFailed() from exc

        return headers



from typing import Tuple
from pydantic import (
    BaseModel, 
    validator, 
    StrictStr, 
    StrictInt, 
    StrictBool
)

class ChannelHookValidator(BaseModel):
    names: Tuple[StrictStr, ...]
    order: StrictInt
    skip: StrictBool

    @validator('names')
    def validate_names(cls, vals):
        assert len(vals) > 0
        assert len(vals) == len(set(vals))

        return vals



import random
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.experiments.mutations.types.base.mutation import Mutation
from hedra.core.experiments.mutations.types.base.mutation_type import MutationType
from typing import Union, Tuple, List
from .validator import InjectHeaderValidator


Request = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action
]


class InjectHeader(Mutation):

    def __init__(
        self, 
        name: str, 
        chance: float,
        *targets: Tuple[str, ...],
        header_name: str=None,
        header_value: Union[str, bytes, bytearray]=None
    ) -> None:
        super().__init__(
            name, 
            chance,
            MutationType.INJECT_HEADER,
            *targets
        )

        validated_mutation = InjectHeaderValidator(
            header_name=header_name,
            header_value=header_value
        )

        self.header_name = validated_mutation.header_name
        self.header_value = validated_mutation.header_value
        self.original_headers: Union[str, List[Tuple[str, str]]]

    async def mutate(self, action: Request=None) -> Request:
        chance_roll = random.uniform(0, 1)
        if chance_roll <= self.chance:
            return action
        
        original_headers = list(action._header_items)

        action._header_items: List[Tuple[str, str]] = original_headers
        mutation_header = (
            self.header_name,
            self.header_value
        )

        if mutation_header not in action._header_items:
            action._header_items.append(mutation_header)
            action._setup_headers()

        action._headers[self.header_name] = self.header_value

        return action
    
    def copy(self):
        return InjectHeader(
            self.name,
            self.chance,
            *list(self.targets),
            header_name=self.header_name,
            header_value=self.header_value
        )


from .http import HTTPClient
from .http2 import HTTP2Client
from .http3 import HTTP3Client
from .grpc import GRPCClient
from .graphql import GraphQLClient
from .graphql_http2 import GraphQLHTTP2Client
from .websocket import WebsocketClient
from .playwright import PlaywrightClient
from .task import TaskClient
from .udp import UDPClient


from .validator import DependsValidator


def depends(*stages):
    DependsValidator(stages=stages)

    def wrapper(cls):

        def decorator():

            direct_dependencies = [stage for stage in stages]
            cls.dependencies = direct_dependencies
            
            return cls

        return decorator()

    return wrapper


from typing import Any, Dict, Optional
from hedra.reporting.types.common.types import ReporterTypes
from pydantic import BaseModel


class KafkaConfig(BaseModel):
    host: str='localhost:9092'
    client_id: str='hedra'
    events_topic: str='events'
    metrics_topic: str='metrics'
    experiments_topic: str='experiments'
    streams_topic: str='streams'
    system_metrics_topic: str='system_metrics'
    events_partition: int=0
    metrics_partition: int=0
    experiments_partition: int=0
    streams_partition: int=0
    system_metrics_partition: int=0
    compression_type: Optional[str]
    timeout: int=1000
    idempotent: bool=True
    options: Dict[str, Any]={}
    reporter_type: ReporterTypes=ReporterTypes.Kafka


from typing import Union
from .metric_types import (
    metric_type_map,
    MetricType
)


class CustomMetric:

    def __init__(
            self, 
            metric_name: str, 
            metric_shortname: str,
            metric_value: Union[float, int],
            metric_group: str=None,
            metric_type: str='sample'
        ) -> None:
        self.metric_shortname: str = metric_shortname
        self.metric_name: str = metric_name
        self.metric_value: Union[float, int] = metric_value
        self.metric_group = metric_group
        self.metric_type = metric_type_map.get(
            metric_type,
            MetricType.SAMPLE
        )


"""Synchronization primitives."""

__all__ = ('Lock', 'Event', 'Condition', 'Semaphore',
           'BoundedSemaphore', 'Barrier')

import collections
from asyncio import exceptions
from asyncio import mixins

class _ContextManagerMixin:
    async def __aenter__(self):
        await self.acquire()
        # We have no use for the "as ..."  clause in the with
        # statement for locks.
        return None

    async def __aexit__(self, exc_type, exc, tb):
        self.release()

class BalancingSemaphore(_ContextManagerMixin, mixins._LoopBoundMixin):

    __slots__ = (
        '_value',
        '_slots',
        '_waiters',
        '_busy',
        '_wakeup_scheduled'
    )

    """A Semaphore implementation.
    A semaphore manages an internal counter which is decremented by each
    acquire() call and incremented by each release() call. The counter
    can never go below zero; when acquire() finds that it is zero, it blocks,
    waiting until some other thread calls release().
    Semaphores also support the context management protocol.
    The optional argument gives the initial value for the internal
    counter; it defaults to 1. If the value given is less than 0,
    ValueError is raised.
    """

    def __init__(self, value=1):
        if value < 0:
            raise ValueError("Semaphore initial value must be >= 0")
        self._value = value
        self._slots = [idx for idx in range(value)]
        self._waiters = tuple([collections.deque() for _ in self._slots])
        self._busy = bytearray([0 for _ in self._slots])
        self._wakeup_scheduled = False

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else f'unlocked, value:{self._value}'
        if self._waiters:
            extra = f'{extra}, waiters:{len(self._waiters)}'
        return f'<{res[1:-1]} [{extra}]>'

    def _wake_up_next(self, min_idx: int):
        while self._waiters[min_idx]:
            waiter = self._waiters[min_idx].popleft()
            if not waiter.done():
                waiter.set_result(None)
                self._wakeup_scheduled = True
                return

    def locked(self):
        """Returns True if semaphore can not be acquired immediately."""
        return self._value == 0

    async def acquire(self):
        """Acquire a semaphore.
        If the internal counter is larger than zero on entry,
        decrement it by one and return True immediately.  If it is
        zero on entry, block, waiting until some other coroutine has
        called release() to make it larger than 0, and then return
        True.
        """
        # _wakeup_scheduled is set if *another* task is scheduled to wakeup
        # but its acquire() is not resumed yet

        # Find the line with the smalles number of futures waiting. Assign
            # to that queue
        min_queue = min(self._busy)
        min_idx = self._busy.index(min_queue)

        # while the line of the minimum is not empty.
        while self._wakeup_scheduled or self._busy[min_idx] > 0:
            fut = self._get_loop().create_future()
            self._waiters[min_idx].append(fut)

            try:
                await fut

                self._wakeup_scheduled = False
            except exceptions.CancelledError:
                self._wake_up_next(min_idx)
                raise
        
        # A line now has waiters.
        self._busy[min_idx] += 1

        return True

    def release(self):
        """Release a semaphore, incrementing the internal counter by one.
        When it was zero on entry and another coroutine is waiting for it to
        become larger than zero again, wake up that coroutine.
        """
        max_queue = max(self._busy)
        max_idx = self._busy.index(max_queue)

        self._busy[max_idx] -= 1
        self._wake_up_next(max_idx)


import asyncio
from ssl import SSLContext
from typing import Tuple, Optional, Union
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.protocols.tcp import TCPConnection
from hedra.core.engines.types.http2.streams.stream_settings import Settings
from hedra.core.engines.types.http2.streams.stream_settings_codes import SettingCodes
from .stream import Stream
from .frames import FrameBuffer
from .frames.types.base_frame import Frame

class HTTP2Connection:

    __slots__ = (
        'timeouts',
        'connected',
        'init_id',
        'reset_connection',
        'stream_type',
        'init_id',
        'stream_id',
        'concurrency',
        'dns_address',
        'port',
        'connection',
        'lock',
        'stream',
        'local_settings',
        'remote_settings',
        'outbound_flow_control_window',
        'local_settings_dict',
        'remote_settings_dict',
        'settings_frame',
        'headers_frame',
        'window_update_frame'
    )

    def __init__(self, stream_id: int, timeouts: Timeouts, concurrency: int, reset_connection: bool, stream_type: RequestTypes) -> None:
        self.timeouts = timeouts
        self.connected = False
        self.init_id = stream_id
        self.reset_connection = reset_connection
        self.stream_type = stream_type

        if self.init_id%2 == 0:
            self.init_id += 1

        self.stream_id = 0
        self.concurrency = concurrency
        self.dns_address = None
        self.port = None

        self.connection = TCPConnection(stream_type)
        self.lock = asyncio.Lock()
        self.stream = Stream(self.stream_id, self.timeouts)

        self.local_settings = Settings(
            client=True,
            initial_values={
                SettingCodes.ENABLE_PUSH: 0,
                SettingCodes.MAX_CONCURRENT_STREAMS: concurrency,
                SettingCodes.MAX_HEADER_LIST_SIZE: 65535,
            }
        )
        self.remote_settings = Settings(
            client=False
        )

        self.outbound_flow_control_window = self.remote_settings.initial_window_size

        del self.local_settings[SettingCodes.ENABLE_CONNECT_PROTOCOL]

        self.local_settings_dict = {setting_name: setting_value for setting_name, setting_value in self.local_settings.items()}
        self.remote_settings_dict = {setting_name: setting_value for setting_name, setting_value in self.remote_settings.items()}

        self.settings_frame = Frame(0, 0x04, settings=self.local_settings_dict)
        self.headers_frame = Frame(self.init_id, 0x01)
        self.headers_frame.flags.add('END_HEADERS')
        
        self.window_update_frame = Frame(self.init_id, 0x08, window_increment=65536)

        self.stream.connection_data.extend(self.settings_frame.serialize())

    async def connect(self, 
        hostname: str, 
        dns_address: str,
        port: int, 
        socket_config: Tuple[int, int, int, int, Tuple[int, int]], 
        ssl: Optional[SSLContext]=None,
        timeout: Optional[float] = None
    ) -> Union[Stream, Exception]:
        
            try:
                if self.connected is False or self.dns_address != dns_address or self.reset_connection:
                    reader, writer = await asyncio.wait_for(
                        self.connection.create_http2(
                            hostname,
                            socket_config=socket_config,
                            ssl=ssl
                        ),
                        timeout=timeout
                    )

                    self.connected = True
                    self.stream_id = self.init_id
                    self.dns_address = dns_address
                    self.port = port

                    self.stream.reader = reader
                    self.stream.writer = writer
                  
                    self.stream.headers_frame = self.headers_frame
                    self.stream.window_frame = self.window_update_frame

                else:

                    self.stream_id += 2# self.concurrency
                    if self.stream_id%2 == 0:
                        self.stream_id += 1

                    self.stream.stream_id = self.stream_id
                    self.stream.headers_frame.stream_id = self.stream_id
                    self.stream.window_frame.stream_id = self.stream_id

                self.stream.frame_buffer = FrameBuffer()
                return self.stream
            
            except asyncio.TimeoutError:
                raise Exception('Connection timed out.')

            except ConnectionResetError:
                raise Exception('Connection reset.')

            except Exception as e:
                raise e

    async def close(self):
        await self.connection.close()


import asyncio
from typing import Generic, TypeVar
from typing import Coroutine, List, Dict


A = TypeVar('A')


class Hooks(Generic[A]):

    __slots__ = (
        'before',
        'after',
        'checks',
        'notify',
        'listen',
        'channel_events',
        'listeners',
        'channels'
    )

    def __init__(
        self,
        before: List[List[Coroutine]] = None,
        after: List[List[Coroutine]] = None,
        checks: List[Coroutine] = None
    ) -> None:
        self.before: List[List[Coroutine]] = before
        self.after: List[List[Coroutine]] = after
        self.checks: List[List[Coroutine]] = checks
        self.notify = False
        self.listen = False
        self.channel_events: List[asyncio.Event] = []
        self.listeners: Dict[str, A] = {}
        self.channels: List[Coroutine] = []

    @property
    def channel_hook_names(self):
        if self.channels:
            return [
                channel.name for channel in self.channels
            ]

        return []

    def to_names(self):

        names = {}

        if self.before:
            names['before'] = self.before_hook_name

        if self.after:
            names['after'] = self.after_hook_name

        check_names = []
        for check in self.checks:
            check_names.append(check.name)

        names['checks'] = check_names
        names['channels'] = self.channel_hook_names

        names['listeners'] = [
            listener_name for listener_name in self.listeners
        ]

        return names

    def action_to_serializable(self):
        return {
            'notify': self.notify,
            'listen': self.listen,
            'names': self.to_names()
        }



from collections import defaultdict
from typing import Dict, List
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.channel.hook import ChannelHook
from hedra.core.hooks.types.check.hook import CheckHook
from hedra.core.hooks.types.condition.hook import ConditionHook
from hedra.core.hooks.types.context.hook import ContextHook
from hedra.core.hooks.types.event.hook import EventHook
from hedra.core.hooks.types.load.hook import LoadHook
from hedra.core.hooks.types.metric.hook import MetricHook
from hedra.core.hooks.types.save.hook import SaveHook
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.hooks.types.transform.hook import TransformHook
from .hook import Hook
from .hook_type import HookType



class Registrar:
    all: Dict[str, List[Hook]] = {}
    reserved: Dict[str, Dict[str, Hook]] = defaultdict(dict)
    module_paths: Dict[str, str] = {}

    def __init__(self, hook_type) -> None:
        self.hook_type = hook_type
        self.hook_types = {
            HookType.ACTION: lambda *args, **kwargs: ActionHook(*args, **kwargs),
            HookType.CHANNEL: lambda *args, **kwargs:  ChannelHook(*args, **kwargs),
            HookType.CHECK: lambda *args, **kwargs:  CheckHook(*args, **kwargs),
            HookType.CONDITION: lambda *args, **kwargs: ConditionHook(*args, **kwargs),
            HookType.CONTEXT: lambda *args, **kwargs:  ContextHook(*args, **kwargs),
            HookType.EVENT: lambda *args, **kwargs:  EventHook(*args, **kwargs),
            HookType.METRIC: lambda *args, **kwargs:  MetricHook(*args, **kwargs),
            HookType.LOAD: lambda *args, **kwargs:  LoadHook(*args, **kwargs),
            HookType.SAVE: lambda *args, **kwargs:  SaveHook(*args, **kwargs),
            HookType.TASK: lambda *args, **kwargs:  TaskHook(*args, **kwargs),
            HookType.TRANSFORM: lambda *args, **kwargs: TransformHook(*args, **kwargs),
        }

    def __call__(self, hook):
        self.module_paths[hook.__name__] = hook.__module__

  
        def wrap_hook(*args, **kwargs):

            def wrapped_method(func):

                hook_name = func.__qualname__
                hook_shortname = func.__name__

                hook = self.hook_types[self.hook_type]

                hook_args = args
                args_count = len(args)
                
                if args_count < 1:
                    hook_args = []
                        
                if hook_name not in self.all:
                    self.all[hook_name] = [
                        hook(
                            hook_name,
                            hook_shortname,
                            func,
                            *hook_args,
                            **kwargs
                        )
                    ]

                else:
                    self.all[hook_name].append(hook(
                        hook_name,
                        hook_shortname,
                        func,
                        *hook_args,
                        **kwargs
                    ))

                return func
            
            return wrapped_method

        return wrap_hook
        

def makeRegistrar():
    return Registrar


registrar = makeRegistrar()


import asyncio
import uuid
import functools
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)

try:
    from prometheus_client.exposition import basic_auth_handler
    from prometheus_client import (
        CollectorRegistry,
        push_to_gateway,
    )
    from .prometheus_metric import PrometheusMetric
    from prometheus_client.core import REGISTRY
    from .prometheus_config import PrometheusConfig
    has_connector = True

except Exception:
    PrometheusConfig = None
    has_connector = False
    basic_auth_handler = lambda: None


class Prometheus:

    def __init__(self, config: PrometheusConfig) -> None:
        self.pushgateway_address = config.pushgateway_address
        self.auth_request_method = config.auth_request_method
        self.auth_request_timeout = config.auth_request_timeout
        self.auth_request_data = config.auth_request_data
        self.username = config.username
        self.password = config.password
        self.namespace = config.namespace
        self.job_name = config.job_name

        self.registry = None
        self._auth_handler = None
        self._has_auth = False
        self._loop = asyncio.get_event_loop()
        
        self.types_map = {
            'total': 'count',
            'succeeded': 'count',
            'failed': 'count',
            'actions_per_second': 'gauge',
            'median': 'gauge',
            'mean': 'gauge',
            'variance': 'gauge',
            'stdev': 'gauge',
            'minimum': 'gauge',
            'maximum': 'gauge'
        }

        self._events: Dict[str, PrometheusMetric] = {}
        self._metrics: Dict[str, Dict[str, Dict[str, PrometheusMetric]]] = {}
        self._streams: Dict[str, Dict[str, PrometheusMetric]] = {}
        self._session_system_metrics: Dict[str, Dict[str, PrometheusMetric]] = {}
        self._stage_system_metrics: Dict[str, Dict[str, PrometheusMetric]] = {}

        self._experiments: Dict[str, Dict[str, PrometheusMetric]] = {}
        self._variants: Dict[str, Dict[str, PrometheusMetric]] = {}
        self._mutations: Dict[str, Dict[str, PrometheusMetric]] = {}

        self._shared_metrics: Dict[str, Dict[str, PrometheusMetric]]  = {}
        self._custom_metrics: Dict[str, PrometheusMetric] = {}
        self._errors: Dict[str, PrometheusMetric] = {}

        self.metric_types_map = {
            MetricType.COUNT: 'count',
            MetricType.RATE: 'gauge',
            MetricType.DISTRIBUTION: 'histogram',
            MetricType.SAMPLE: 'gauge'
        }

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()
        
    async def connect(self) -> None:

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Prometheus Pushgateway at: {self.pushgateway_address}')

        self.registry = CollectorRegistry()
        REGISTRY.register(self.registry)

        if self.username and self.password:
            self._has_auth = True

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Prometheus Pushgateway at: {self.pushgateway_address}')
            

    def _generate_auth(self) -> basic_auth_handler:
        return basic_auth_handler(
            self.pushgateway_address,
            self.auth_request_method,
            self.auth_request_timeout,
            {
                'Content-Type': 'application/json'
            },
            self.auth_request_data,
            username=self.username,
            password=self.password
        )
    
    async def _submit_to_pushgateway(self):
        if self._has_auth:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushing to secure Prometheus Pushgateway via HTTPS')
            await self._loop.run_in_executor(
                None,
                functools.partial(
                    push_to_gateway,
                    self.pushgateway_address,
                    job=self.job_name,
                    registry=self.registry,
                    handler=self._generate_auth
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushed to secure Prometheus Pushgateway via HTTPS')

        else:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushing to Prometheus Pushgateway via HTTP')
            await self._loop.run_in_executor(
                None,
                functools.partial(
                    push_to_gateway,
                    self.pushgateway_address,
                    job=self.job_name,
                    registry=self.registry
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushed to Prometheus Pushgateway via HTTP')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        for quantile in SystemMetricsSet.quantiles:
            quantile_key = f'quantile_{quantile}th'
            self.types_map[quantile_key] = 'gauge'

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Prometheus - Namespace: {self.namespace}')

        metrics_sets: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.name}:{metrics_set.group}')

            session_system_metric_set = self._session_system_metrics.get(metrics_set.name)
            if session_system_metric_set is None:
                tags = [
                    f'name:{metrics_set.name}',
                    f'group:{metrics_set.group}'
                ]


                fields = {}

                for metric_field in metrics_set.record.keys():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics Group - {metrics_set.name}:{metrics_set.group}:{metric_field}')
                    metric_type = self.types_map.get(metric_field)
                    metric_name = f'{metrics_set.name}_{metrics_set.group}_{metric_field}'.replace('.', '_')

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{metrics_set.name} {metrics_set.group} {metric_field}',
                        metric_labels=tags,
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()

                    fields[metric_field] = prometheus_metric

                self._session_system_metrics[metrics_set.name] = fields
            
            for field, prometheus_metric in session_system_metric_set.items():
                metric_value = metrics_set.record.get(field)
                prometheus_metric.update(metric_value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to Prometheus - Namespace: {self.namespace}')
    
    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        for quantile in SystemMetricsSet.quantiles:
            quantile_key = f'quantile_{quantile}th'
            self.types_map[quantile_key] = 'gauge'

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Prometheus - Namespace: {self.namespace}')

        metrics_sets: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')
            
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.name}:{metrics_set.group}')

            stage_system_metric_set = self._stage_system_metrics.get(metrics_set.name)
            if stage_system_metric_set is None:
                tags = [
                    f'name:{metrics_set.name}',
                    f'group:{metrics_set.group}'
                ]


                fields = {}

                for metric_field in metrics_set.record.keys():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics Group - {metrics_set.name}:{metrics_set.group}:{metric_field}')
                    metric_type = self.types_map.get(metric_field)
                    metric_name = f'{metrics_set.name}_{metrics_set.group}_{metric_field}'.replace('.', '_')

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{metrics_set.name} {metrics_set.group} {metric_field}',
                        metric_labels=tags,
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()

                    fields[metric_field] = prometheus_metric

                self._stage_system_metrics[metrics_set.name] = fields
            
            for field, prometheus_metric in stage_system_metric_set.items():
                metric_value = metrics_set.record.get(field)
                prometheus_metric.update(metric_value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to Prometheus - Namespace: {self.namespace}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Prometheus - Namespace: {self.namespace}')

        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{stream.stream_set_id}')

            stream_metric_name = f'{stage_name}_streams'

            stream_metric_set = self._streams.get(stage_name)
            if stream_metric_set is None:
                stream_metric_set = {}

                for group_name, group in stream.grouped.items():
                    group_metrics = stream_metric_set.get(group_name)

                    tags = [
                        f'name:{stage_name}_streams',
                        f'stage:{stage_name}',
                        f'group:{group_name}'
                    ]

                    if group_metrics is None:
                        group_metrics = {}

                        for quantile in stream.quantiles:
                            quantile_key = f'quantile_{quantile}th'
                            self.types_map[quantile_key] = 'gauge'

                        fields = {}

                        for metric_field in group.keys():
                            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream Group - {stage_name}:{group_name}:{metric_field}')
                            metric_type = self.types_map.get(metric_field)
                            metric_name = f'{stage_name}_{group_name}_{metric_field}'.replace('.', '_')

                            prometheus_metric = PrometheusMetric(
                                metric_name,
                                metric_type,
                                metric_description=f'{stream_metric_name} {metric_field}',
                                metric_labels=tags,
                                metric_namespace=self.namespace,
                                registry=self.registry
                            )
                            prometheus_metric.create_metric()

                            fields[metric_field] = prometheus_metric

                        group_metrics[group_name] = fields

                stream_metric_set.update(group_metrics)

            self._streams[stage_name] = stream_metric_set

            for group_name, group in stream.grouped.items():
                group_metrics = stream_metric_set.get(group_name)

                for field in group_metrics[group_name]:
                    metric_value = group.get(field)
                    group_metrics.update(metric_value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to Prometheus - Namespace: {self.namespace}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Prometheus - Namespace: {self.namespace}')
        
        for experiment in experiment_metrics.experiment_summaries:
            
            experiment_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiments Set - {experiment.experiment_name}:{experiment_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in experiment.tags
            ]
            
            experiment_stats = self._experiments.get(experiment.experiment_name)
            if experiment_stats is None:
                experiment_stats = {}

                for field in experiment.stats.keys():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiment- {experiment.experiment_name}:{field}')

                    metric_name = f'{experiment.experiment_name}_{field}'.replace('.', '_')
                    metric_type = self.types_map.get(field)

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{experiment.experiment_name} {field}',
                        metric_labels=[
                            *tags
                        ],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()
                    
                    experiment_stats[field] = prometheus_metric
                
                self._experiments[experiment.experiment_name] = experiment_stats

            self._experiments[experiment.experiment_name] = experiment_stats
            for field, value in experiment.stats.items():
                metric = experiment_stats.get(field)
                metric.update(value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Prometheus - Namespace: {self.namespace}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Prometheus - Namespace: {self.namespace}')
        
        for variant in experiment_metrics.variant_summaries:
            
            variant_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variants Set - {variant.variant_name}:{variant_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in variant.tags
            ]
            
            variant_stats = self._variants.get(variant.variant_name)
            if variant_stats is None:
                variant_stats = {}

                for field in variant.stats.keys():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variants- {variant.variant_name}:{field}')

                    metric_name = f'{variant.variant_name}_{field}'.replace('.', '_')
                    metric_type = self.types_map.get(field)

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{variant.variant_name} {field}',
                        metric_labels=[
                            *tags
                        ],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()
                    
                    variant_stats[field] = prometheus_metric
                
                self._variants[variant.variant_name] = variant_stats

            self._variants[variant.variant_name] = variant_stats
            for field, value in variant.stats.items():
                metric = variant_stats.get(field)
                metric.update(value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Prometheus - Namespace: {self.namespace}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Prometheus - Namespace: {self.namespace}')
        
        for mutation in experiment_metrics.mutation_summaries:
            
            mutation_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutations Set - {mutation.mutation_name}:{mutation_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in mutation.tags
            ]
            
            mutation_stats = self._mutations.get(mutation.mutation_name)
            if mutation_stats is None:
                mutation_stats = {}

                for field in mutation.stats.keys():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutations- {mutation.mutation_name}:{field}')

                    metric_name = f'{mutation.mutation_name}_{field}'.replace('.', '_')
                    metric_type = self.types_map.get(field)

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{mutation.mutation_name} {field}',
                        metric_labels=[
                            *tags
                        ],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()
                    
                    mutation_stats[field] = prometheus_metric
                
                self._mutations[mutation.mutation_name] = mutation_stats

            self._mutations[mutation.mutation_name] = mutation_stats
            for field, value in mutation.stats.items():
                metric = mutation_stats.get(field)
                metric.update(value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Prometheus - Namespace: {self.namespace}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Prometheus - Namespace: {self.namespace}')

        for event in events:
            
            record = {
                f'{event.name}_time': event.time,
                f'{event.name}_success': 1 if event.success else 0,
                f'{event.name}_failed': 1 if event.success is False else 0
            }

            if self._events.get(event.name) is None:

                fields = {}

                for event_field in record.keys():
                    metric_type = self.types_map.get(event_field)
                
                    metric = PrometheusMetric(
                        f'{event.name}_{event_field}',
                        metric_type,
                        metric_description=f'{event.name} {event_field}',
                        metric_labels=[],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )

                    metric.create_metric()

                    fields[event_field] = metric

                self._events[event.name] = fields
                
            for event_field, event_value in record.items():
                if event_value and event_field in self.types_map:
                    self._events[event.name][event_field].update(event_value)

            await self._submit_to_pushgateway()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Prometheus - Namespace: {self.namespace}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Prometheus - Namespace: {self.namespace}')
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]
            
            shared_metrics = self._shared_metrics.get(metrics_set.name)
            if shared_metrics is None:
                shared_metrics = {}

                for field in metrics_set.common_stats.keys():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')

                    metric_name = f'{metrics_set.name}_{field}'.replace('.', '_')
                    metric_type = self.types_map.get(field)

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{metrics_set.name} {field}',
                        metric_labels=[
                            *tags,
                            f'stage:{metrics_set.stage}',
                            'group:common'
                        ],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()
                    
                    shared_metrics[field] = prometheus_metric

            self._shared_metrics[metrics_set.name] = shared_metrics
            for field, value in metrics_set.common_stats.items():
                metric = shared_metrics.get(field)
                metric.update(value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Prometheus - Namespace: {self.namespace}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Prometheus - Namespace: {self.namespace}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]

            stage_metrics = self._metrics.get(metrics_set.name)
            if stage_metrics is None:
                stage_metrics = {}

                for group_name, group in metrics_set.groups.items():
                    group_metrics = stage_metrics.get(group_name)

                    if group_metrics is None:
                        group_metrics = {}

                        for quantile_name in metrics_set.quantiles:
                            self.types_map[quantile_name] = 'gauge'

                        fields = {}

                        for metric_field in metrics_set.stats_fields:
                            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{metric_field}')
                            metric_type = self.types_map.get(metric_field)
                            metric_name = f'{metrics_set.name}_{group_name}_{metric_field}'.replace('.', '_')

                            prometheus_metric = PrometheusMetric(
                                metric_name,
                                metric_type,
                                metric_description=f'{metrics_set.name} {metric_field}',
                                metric_labels=[
                                    *tags,
                                    f'stage:{metrics_set.stage}',
                                    f'group:{group_name}'
                                ],
                                metric_namespace=self.namespace,
                                registry=self.registry
                            )
                            prometheus_metric.create_metric()

                            fields[metric_field] = prometheus_metric

                        group_metrics[group_name] = fields

                stage_metrics[metrics_set.name] = group_metrics

            self._metrics[metrics_set.name] = stage_metrics
            for group_name, group in metrics_set.groups.items():
                group_metrics = stage_metrics.get(group_name)
                record = group.record

                for field in group_metrics[group_name]:
                    metric_value = record.get(field)
                    field_metric = group_metrics.get(field)
                    field_metric.update(metric_value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Prometheus - Namespace: {self.namespace}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Prometheus - Namespace: {self.namespace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            custom_metrics = self._custom_metrics.get(metrics_set.name)
            if custom_metrics is None:
                custom_metrics = {}

                for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():
                    # group_metrics = custom_metrics.get(custom_group_name)

                    metric_type = self.metric_types_map.get(
                        custom_metric.metric_type,
                        'gauge'
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:custom:{custom_metric_name}')

                    metric_name = f'{metrics_set.name}_{custom_metric_name}'.replace('.', '_')

                    tags = [
                        f'{tag.name}:{tag.value}' for tag in metrics_set.tags
                    ]

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{metrics_set.name} {custom_metric_name}',
                        metric_labels=[
                            *tags,
                            f'stage:{metrics_set.stage}',
                            'group:custom',
                            f'type:{custom_metric.metric_type}'
                        ],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()

                    custom_metrics[custom_metric_name] = prometheus_metric


            self._custom_metrics[metrics_set.name] = custom_metrics
            for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():
                custom_metric_prometheus_value: PrometheusMetric = custom_metrics.get(custom_metric_name)
                custom_metric_prometheus_value.update(custom_metric.metric_value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Prometheus - Namespace: {self.namespace}')
                    
    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Prometheus - Namespace: {self.namespace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._errors.get(metrics_set.name) is None:
                errors_metric = PrometheusMetric(
                    f'{metrics_set.name}_errors',
                    'count',
                    metric_description=f'Errors for action - {metrics_set.name}.',
                    metric_labels=[],
                    metric_namespace=self.namespace,
                    registry=self.registry
                )

                errors_metric.create_metric()
                self._errors[metrics_set.name] = errors_metric

            for error in metrics_set.errors:              
                errors_metric.update(
                    error.get('count'),
                    labels={
                        'message': error.get('message')
                    }
                )
                
        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Prometheus - Namespace: {self.namespace}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
                


from __future__ import annotations
import uuid
import numpy
from collections import defaultdict
from typing import Any, Dict, Union
from hedra.reporting.stats import (
    Median,
    Mean,
    Variance,
    StandardDeviation
)
from .results import results_types
from .types.task_processed_result import TaskProcessedResult
from .types.base_processed_result import BaseProcessedResult


def default_count():
    return 0


class ProcessedResultsGroup:

    __slots__ = (
        'events_group_id',
        'source',
        'groups',
        'events',
        'timings',
        'tags',
        'total',
        'succeeded',
        'failed',
        'errors',
        '_streaming_mean',
        '_streaming_variance',
        '_streaming_stdev',
        '_streaming_median'
    )

    def __init__(self) -> None:

        self.events_group_id = str(uuid.uuid4())

        self.groups: Dict[Dict[str, Union[int, float]]] = {}
        self.timings = defaultdict(list)
        self.source = None
        self.tags = {}
        self.total = 0
        self.succeeded = 0
        self.failed = 0
        self.errors = defaultdict(default_count)

        self._streaming_mean = defaultdict(Mean)
        self._streaming_variance = defaultdict(Variance)
        self._streaming_stdev = defaultdict(StandardDeviation)
        self._streaming_median = defaultdict(Median)

    def add(
        self, 
        stage_name: str,
        result: Any, 
    ):

        processed_result: BaseProcessedResult = results_types.get(result.type, TaskProcessedResult)(
            stage_name,
            result
        )

        if self.source is None:
            self.source = processed_result.source

        self.tags.update(processed_result.tags_to_dict())

        if processed_result.error is None:
            self.succeeded += 1
        
        else:
            self.errors[processed_result.error] += 1
            self.failed += 1

        for timing_group, timing in processed_result.timings.items():
            if timing > 0:
                self.timings[timing_group].append(timing)

    def calculate_stats(self):

        self.total = self.succeeded + self.failed

        for group_name, group_timings in self.timings.items():

            if len(group_timings) == 0:
                group_timings = [0]

            median = float((numpy.median(group_timings)))
            mean = float(numpy.mean(group_timings, dtype=numpy.float64))
            variance = float(numpy.var(group_timings, dtype=numpy.float64, ddof=1))
            stdev = float(numpy.std(group_timings, dtype=numpy.float64))

            self._streaming_median[group_name].update(median)
            self._streaming_mean[group_name].update(mean)
            self._streaming_variance[group_name].update(variance)
            self._streaming_stdev[group_name].update(stdev)

            self.groups[group_name] = {
                group_name: {
                    'median': median,
                    'mean': mean,
                    'variance': variance,
                    'stdev': stdev,
                    'minimum': min(group_timings),
                    'maximum': max(group_timings)
                }

            }

    def calculate_quantiles(self):

        quantile_ranges = [ 
            .10, 
            .20, 
            .25, 
            .30, 
            .40, 
            .50, 
            .60, 
            .70, 
            .75, 
            .80, 
            .90, 
            .95, 
            .99 
        ]

        for group_name, group_timings in self.timings.items():

            if len(group_timings) == 0:
                group_timings = [0]

            quantiles = {
                f'quantile_{int(quantile_range * 100)}th': quantile for quantile, quantile_range in zip(
                    numpy.quantile(
                        numpy.array(group_timings), quantile_ranges),
                    quantile_ranges
                )
            }
            
            self.groups[group_name]['quantiles'] = quantiles


from typing import Union
from scipy.stats import f
from .base import BaseDistribution


class FDistribution(BaseDistribution):

    def __init__(
        self,
        size: int,
        distribution_function_number: Union[int, float]=10,
        distribution_function_density: Union[int, float]=6,
        center: Union[int, float]=0.5,
        randomness: Union[int, float]=0.25
    ):
        super().__init__(
            size=size,
            center=center,
            randomness=randomness,
            frozen_distribution=f(
                distribution_function_number,
                distribution_function_density,
                loc=center,
                scale=randomness
            )
        )



from __future__ import annotations
import asyncio
from typing import Optional, Tuple
from hedra.core.engines.types.common.protocols import UDPConnection
from hedra.core.engines.types.common.protocols.udp.quic_protocol import QuicProtocol


class HTTP3Connection:

    __slots__ = (
        'dns_address',
        'port',
        'ip_addr',
        'lock',
        'protocol',
        'connected',
        'reset_connection',
        'pending',
        '_connection_factory'
    )

    def __init__(self, reset_connection: bool=False) -> None:
        self.dns_address: str = None
        self.port: int = None
        self.ip_addr = None
        self.lock = asyncio.Lock()

        self.protocol: QuicProtocol = None
        self.connected = False
        self.reset_connection = reset_connection
        self.pending = 0
        self._connection_factory = UDPConnection()

    async def make_connection(
        self, 
        dns_address: str,
        port: int, 
        socket_config: Tuple[int, int, int, int, Tuple[int, int]],
        server_name: str=None,
        timeout: Optional[float]=None
    ) -> None:
    
        if self.connected is False or self.dns_address != dns_address or self.reset_connection:
            try:
                self.protocol = await asyncio.wait_for(
                    self._connection_factory.create_http3(
                        socket_config=socket_config,
                        server_name=server_name
                    ),
                     
                    timeout=timeout
                )
                    
                self.connected = True

                self.dns_address = dns_address
                self.port = port

            except asyncio.TimeoutError:
                raise Exception('Connection timed out.')

            except ConnectionResetError:
                raise Exception('Connection reset.')

            except Exception as e:
                raise e

    async def close(self):
        if self.protocol:
            self.protocol.close()



from collections import defaultdict
from hedra.core.hooks.types.channel.hook import ChannelHook
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.event import BaseEvent
from hedra.core.hooks.types.base.event_types import EventType


class ChannelEvent(BaseEvent[ChannelHook]):

    def __init__(self, target: Hook, source: ChannelHook) -> None:
        super(
            ChannelEvent,
            self
        ).__init__(
            target,
            source
        )

        self.event_type = EventType.CHANNEL

    def copy(self):
        channel_event = ChannelEvent(
            self.target.copy(),
            self.source.copy()
        )

        channel_event.execution_path = list(self.execution_path)
        channel_event.previous_map = list(self.previous_map)
        channel_event.next_map = list(self.next_map)
        channel_event.next_args = defaultdict(dict)

        return channel_event


from ssl import SSLContext
from typing import List, Optional
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class CassandraConfig(BaseModel):
    hosts: List[str] = ['127.0.0.1']
    port: int=9042
    username: Optional[str]=None
    password: Optional[str]=None
    keyspace: str='hedra'
    events_table: str='events'
    metrics_table: str='metrics'
    streams_table: str='streams'
    experiments_table: str='experiments'
    system_metrics_table: str='system_metrics'
    replication_strategy: str='SimpleStrategy'
    replication: int=3
    ssl: Optional[SSLContext]=None
    reporter_type: ReporterTypes=ReporterTypes.Cassandra

    class Config:
        arbitrary_types_allowed = True


import asyncio
import functools
import collections
import collections.abc
import uuid
import psutil
import signal
import os
import time
from pathlib import Path
from typing import List, TextIO, Dict, Union
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from xml.dom.minidom import parseString
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric.metrics_set import MetricsSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from .xml_config import XMLConfig

try:
    from dicttoxml import dicttoxml
except Exception:
    dicttoxml = object

collections.Iterable = collections.abc.Iterable


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop, 
    events_file: TextIO
): 
    try:
        events_file.close()
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


MetricRecord = Dict[str, Union[int, float, str]]
MetricRecordGroup = Dict[str, List[MetricRecord]]
MetricRecordCollection = Dict[str, MetricRecord]
    

class XML:
    
    def __init__(self, config: XMLConfig) -> None:
        self.events_filepath = config.events_filepath
        self.metrics_filepath = Path(config.metrics_filepath).absolute()
        self.experiments_filepath = config.experiments_filepath
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop: asyncio.AbstractEventLoop = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        experiments_path = Path(self.experiments_filepath)
        experiments_directory = experiments_path.parent
        experiments_filename = experiments_path.stem

        self.variants_filepath = os.path.join(
            experiments_directory,
            f'{experiments_filename}_variants.xml'
        )

        self.mutations_filepath = os.path.join(
            experiments_directory,
            f'{experiments_filename}_mutations.xml'
        )

        filepath = Path(config.metrics_filepath)
        base_filepath = filepath.parent
        base_filename = filepath.stem

        self.shared_metrics_filepath = os.path.join(
            base_filepath,
            f'{base_filename}_shared.xml'
        )

        self.custom_metrics_filepath = os.path.join(
            base_filepath,
            f'{base_filename}_custom.xml'
        )

        self.errors_metrics_filepath = os.path.join(
            base_filepath,
            f'{base_filename}_errors.xml'
        )

        self.streams_metrics_filepath = config.streams_filepath

        system_metrics_path = Path(config.system_metrics_filepath)
        system_metrics_directory = system_metrics_path.parent
        system_metrics_filename = system_metrics_path.stem

        self.stage_system_metrics_filepath = os.path.join(
            system_metrics_directory,
            f'{system_metrics_filename}_stages.xml'
        )

        self.session_system_metrics_filepath = os.path.join(
            system_metrics_directory,
            f'{system_metrics_filename}_session.xml'
        )

        self.events_file: TextIO = None
        self.metrics_file: TextIO = None
        self.experiments_file: TextIO = None
        self.variants_file: TextIO = None
        self.mutations_file: TextIO = None
        self.streams_file: TextIO = None
        self.stage_system_metrics_file: TextIO = None
        self.session_system_metrics_file: TextIO = None

        self.write_mode = 'w' if config.overwrite else 'a'

    async def connect(self):
        self._loop = asyncio._get_running_loop()
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Setting filepaths')
        
        original_filepath = Path(self.events_filepath)
        
        directory = original_filepath.parent
        filename = original_filepath.stem

        events_file_timestamp = time.time()

        self.events_filepath = os.path.join(
            directory,
            f'{filename}_{events_file_timestamp}.xml'
        )

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        if self.session_system_metrics_file is None:
            self.session_system_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.session_system_metrics_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.session_system_metrics_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Session System Metrics to file - {self.session_system_metrics_filepath}')

        metrics_sets: Dict[str, MetricRecord] = {}

        for metrics_set in system_metrics_sets:

            for monitor_name, monitor_metrics in metrics_set.session_cpu_metrics.items():

                system_metrics_collection: MetricRecordCollection = {}

                memory_metrics = metrics_set.session_memory_metrics.get(monitor_name)

                system_metrics_collection['cpu'] = monitor_metrics.record
                system_metrics_collection['memory'] = memory_metrics.record
                
                metrics_sets[monitor_name] = system_metrics_collection

        system_metrics_xml = dicttoxml(
            metrics_sets, 
            custom_root='system'
        )

        system_metrics_xml = parseString(system_metrics_xml)

        await self._loop.run_in_executor(
            self._executor,
            self.session_system_metrics_file.write,
            system_metrics_xml.toprettyxml()
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Session System Metrics to file - {self.session_system_metrics_filepath}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        if self.stage_system_metrics_file is None:
            self.stage_system_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.stage_system_metrics_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.stage_system_metrics_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Stage System Metrics to file - {self.stage_system_metrics_file}')

        metrics_sets: List[MetricRecordGroup]= []

        for metrics_set in system_metrics_sets:

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                stage_system_metrics_record: MetricRecordGroup = {
                    'cpu': [],
                    'memory': [],
                    'mb_per_vu': []
                }

                for monitor_metrics in stage_cpu_metrics.values():
                    stage_system_metrics_record['cpu'].append(monitor_metrics.record)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    stage_system_metrics_record['memory'].append(monitor_metrics.record)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    stage_system_metrics_record['mb_per_vu'].append(stage_mb_per_vu_metrics.record)

                metrics_sets.append(stage_system_metrics_record)

        system_metrics_xml = dicttoxml(
            metrics_sets, 
            custom_root='system'
        )

        system_metrics_xml = parseString(system_metrics_xml)

        await self._loop.run_in_executor(
            self._executor,
            self.stage_system_metrics_file.write,
            system_metrics_xml.toprettyxml()
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Stage System Metrics to file - {self.stage_system_metrics_filepath}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):
        if self.streams_file is None:
            self.streams_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.streams_metrics_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.streams_file
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Streams to file - {self.streams_metrics_filepath}')

        streams_data = [
            {
                'stage': stream_name,
                **stream_set.grouped 
            } for stream_name, stream_set in stream_metrics.items()
        ]

        streams_xml = dicttoxml(
            streams_data, 
            custom_root='streams'
        )

        streams_xml = parseString(streams_xml)

        await self._loop.run_in_executor(
            self._executor,
            self.streams_file.write,
            streams_xml.toprettyxml()
        )
    
    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Experiments to file - {self.experiments_filepath}')

        if self.experiments_file is None:
            self.experiments_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.experiments_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.experiments_file
                    )
                )

        experiments_xml = dicttoxml(
            experiment_metrics.experiments, 
            custom_root='experiments'
        )

        experiments_xml = parseString(experiments_xml)

        await self._loop.run_in_executor(
            self._executor,
            self.experiments_file.write,
            experiments_xml.toprettyxml()
        )

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        if self.variants_file is None:
            self.variants_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.variants_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.variants_file
                    )
                )

        variants_xml = dicttoxml(
            experiment_metrics.variants, 
            custom_root='variants'
        )

        variants_xml = parseString(variants_xml)

        await self._loop.run_in_executor(
            self._executor,
            self.variants_file.write,
            variants_xml.toprettyxml()
        )

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        if self.mutations_file is None:
            self.mutations_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.mutations_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.mutations_file
                    )
                ) 

        

        mutations_xml = dicttoxml(
            experiment_metrics.mutations, 
            custom_root='mutations'
        )
        
        mutations_xml = parseString(mutations_xml)

        await self._loop.run_in_executor(
            self._executor,
            self.mutations_file.write,
            mutations_xml.toprettyxml()
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Events to file - {self.events_filepath}')

    async def connect(self):
        self._loop = asyncio._get_running_loop()
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping connect')
        
        original_filepath = Path(self.events_filepath)
        
        directory = original_filepath.parent
        filename = original_filepath.stem

        events_file_timestamp = time.time()

        self.events_filepath = os.path.join(
            directory,
            f'{filename}_{events_file_timestamp}.xml'
        )

    async def submit_events(self, events: List[BaseProcessedResult]):

        if self.events_file is None:
            self.events_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.events_filepath,
                    self.write_mode
                )
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                self._loop.add_signal_handler(
                    getattr(signal, signame),
                    lambda signame=signame: handle_loop_stop(
                        signame,
                        self._executor,
                        self._loop,
                        self.events_file
                    )
                )

        events_xml = dicttoxml([
            event.to_dict() for event in events
        ], custom_root='events')

        events_xml = parseString(events_xml)

        await self._loop.run_in_executor(
            self._executor,
            self.events_file.write,
            events_xml.toprettyxml()
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Events to file - {self.events_filepath}')

    async def submit_common(self, metrics: List[MetricsSet]):
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Shared Metrics to file - {self.metrics_filepath}')

        shared_metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    self.shared_metrics_filepath,
                    'w'
                )
        )

        common_metrics_xml = dicttoxml([
            {
                'name': metric_set.name,
                'stage': metric_set.stage,
                'group': 'common',
                **metric_set.common_stats
            } for metric_set in metrics
        ], custom_root='common_metrics')

        common_metrics_xml = parseString(common_metrics_xml)

        await self._loop.run_in_executor(
            self._executor,
            shared_metrics_file.write,
            common_metrics_xml.toprettyxml()
        )

        await self._loop.run_in_executor(
            self._executor,
            shared_metrics_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Shared Metrics to file - {self.metrics_filepath}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Metrics to file - {self.metrics_filepath}')

        metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                self.metrics_filepath, 
                'w'
            )
        )

        metrics_data = []
        for metrics_set in metrics:
            for group_name, group in metrics_set.groups.items():
                metrics_data.append({
                    **group.record,
                    'group': group_name
                })


        metrics_xml = dicttoxml(metrics_data, custom_root='metrics')
        metrics_xml = parseString(metrics_xml)

        await self._loop.run_in_executor(
            self._executor,
            metrics_file.write,
            metrics_xml.toprettyxml()
        )

        await self._loop.run_in_executor(
            self._executor,
            metrics_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Metrics to file - {self.metrics_filepath}')

    async def submit_custom(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Custom Metrics to file - {self.metrics_filepath}')

        custom_metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                self.custom_metrics_filepath,
                'w'
            )       
        )

        metrics_data = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - Custom')

            metrics_data.append({
                **{
                    cusom_metric_name: custom_metric.metric_value for cusom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                },
                'group': 'custom'
            })

        custom_metrics_xml = dicttoxml(metrics_data, custom_root='custom_metrics')
        custom_metrics_xml = parseString(custom_metrics_xml)

        await self._loop.run_in_executor(
            self._executor,
            custom_metrics_file.write,
            custom_metrics_xml.toprettyxml()
        )

        await self._loop.run_in_executor(
            self._executor,
            custom_metrics_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Custom Metrics to file - {self.metrics_filepath}')


    async def submit_errors(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Error Metrics to file - {self.metrics_filepath}')
        
        errors_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                self.errors_metrics_filepath,
                'w'
            )
        )

        errors = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                errors.append({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                })

        errors_xml = dicttoxml(errors, custom_root='errors')
        errors_xml = parseString(errors_xml)

        await self._loop.run_in_executor(
            self._executor,
            errors_file.write,
            errors_xml.toprettyxml()
        )

        await self._loop.run_in_executor(
            self._executor,
            errors_file.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Error Metrics to file - {self.metrics_filepath}')

    async def close(self):

        if self.events_file:
            await self._loop.run_in_executor(
                self._executor,
                self.events_file.close
            )

        if self.experiments_file:
            await self._loop.run_in_executor(
                self._executor,
                self.experiments_file.close
            )

        if self.variants_file:
            await self._loop.run_in_executor(
                self._executor,
                self.variants_file.close
            )

        if self.mutations_file:
            await self._loop.run_in_executor(
                self._executor,
                self.mutations_file.close
            )
        
        if self.streams_file:
            await self._loop.run_in_executor(
                self._executor,
                self.streams_file.close
            )

        if self.stage_system_metrics_file:
            await self._loop.run_in_executor(
                self._executor,
                self.stage_system_metrics_file.close
            )

        if self.session_system_metrics_file:
            await self._loop.run_in_executor(
                self._executor,
                self.session_system_metrics_file.close
            )

        self._executor.shutdown(cancel_futures=True)
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')



import plotille
from collections import (
    OrderedDict,
    defaultdict
)
from tabulate import tabulate
from typing import (
    List, 
    Union, 
    Dict, 
    Tuple
)
from .table_types import ExecutionResults
from hedra.logging import HedraLogger

CompletionRateSet = Tuple[str, List[Union[int, float]]]


class ExecutionSummaryTable:

    def __init__(
        self,
        execution_results: ExecutionResults
    ) -> None:
        
        self.session_table: Union[str, None] = None
        self.stages_table: Union[str, None] = None
        self.stage_timings_table: Union[str, None] = None

        self.execution_results = execution_results
        self.session_table_rows: List[OrderedDict] = []

        self.stage_summary_tables = defaultdict(list)
        self.stage_streamed_data = defaultdict(dict)
        self._has_streamed = False
        self._graph_time_steps: List[int] = []
        self.actions_and_tasks_common_table_rows: List[OrderedDict] = []
        self.actions_and_tasks_common_table: Union[str, None] = None

        self.session_metrics: List[str] = [
            'total',
            'succeeded',
            'failed'
        ]

        self.logger = HedraLogger()
        self.logger.initialize()

        self.enabled_tables = {
            'session': False,
            'stages': True,
            'actions': True,
            'timings': False
        }

        self.actions_and_tasks_table_rows: Dict[str, List[OrderedDict]] = defaultdict(list)
        self.actions_and_tasks_tables: Dict[str, str] = {}

    def generate_tables(self):
        self._generate_stage_and_session_tables()
        self._generate_actions_and_tasks_table()

        self.session_table = tabulate(
            self.session_table_rows,
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt=".2f"
        )

        self.stages_table = tabulate(
            list(sorted(
                self.stage_summary_tables.get('stage_results'),
                key=lambda row: row['name']
            )),
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt=(
                '.2f', 
                '.2f', 
                '.2f', 
                '.2f', 
                '.2f', 
                '.2f', 
                '.2f', 
                '.2f', 
                '.2E', 
                '.2E', 
                '.2E', 
                '.2E', 
                '.2E'
            )
        )

        for table_name, table_rows in self.actions_and_tasks_table_rows.items():
            self.actions_and_tasks_table_rows[table_name] = list(sorted(
                table_rows,
                key=lambda row: row['name']
            ))

            self.actions_and_tasks_tables[table_name] = tabulate(
                self.actions_and_tasks_table_rows[table_name],
                headers='keys',
                missingval='None',
                tablefmt="simple",
                floatfmt=(
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E', 
                    '.2E',
                    '.2E', 
                    '.2E'
                )
            )

        self.actions_and_tasks_common_table_rows = list(sorted(
            self.actions_and_tasks_common_table_rows,
            key=lambda row: row['name']
        ))

        self.actions_and_tasks_common_table = tabulate(
                self.actions_and_tasks_common_table_rows,
                headers='keys',
                missingval='None',
                tablefmt="simple",
                floatfmt=(
                    '.2f',
                    '.2f',
                    '.2f',
                    '.2f',
                    '.2f',
                    '.2f'
                )
            )

    def show_tables(self):

        self.logger.console.sync.info('')

        if self._has_streamed and self.enabled_tables.get('stages'):
            completion_rates: List[CompletionRateSet] = list(sorted(
                self.stage_streamed_data.get(
                    'completion_rates'
                ).items(),
                key=lambda completion_rate: completion_rate[0]
            ))
            
            for stage_name, stage_completion_rates in completion_rates:

                stage_summary = self.execution_results.get(stage_name)

                graph_time_steps_count = len(self._graph_time_steps)
                stage_completion_rates_count = len(stage_completion_rates)

                graph_size = min(graph_time_steps_count, stage_completion_rates_count)

                scatter_plot = plotille.scatter(
                    self._graph_time_steps[:graph_size],
                    stage_completion_rates[:graph_size],
                    width=120,
                    height=10,
                    y_min=0,
                    x_min=0,
                    x_max=int(round(
                        stage_summary.stage_metrics.time,
                        0
                    )),
                    linesep='\n',
                    X_label='time (sec)',
                    Y_label='completion rate',
                    lc='cyan',
                    marker='⨯'
                )

                self.logger.console.sync.info(f'''\n{stage_name} Completion Rates\n''')
                self.logger.console.sync.info(f'''{scatter_plot}\n''')
        
        if self.enabled_tables.get('session'):
            self.logger.console.sync.info('\n-- Session --\n')
            self.logger.console.sync.info(f'''{self.session_table}\n''')


        if self.enabled_tables.get('stages'):
            self.logger.console.sync.info('\n-- Stages --\n')
            self.logger.console.sync.info(f'''{self.stages_table}\n''')

        if self.enabled_tables.get('actions'):
            self.logger.console.sync.info('\n-- Actions and Tasks --\n')

            self.logger.console.sync.info('Overall:\n')
            self.logger.console.sync.info(f'''{self.actions_and_tasks_common_table}\n''')

        if self.enabled_tables.get('actions') and self.enabled_tables.get('timings'):
            for table_name, table in self.actions_and_tasks_tables.items():
                self.logger.console.sync.info(f'{table_name.capitalize()}:\n')
                self.logger.console.sync.info(f'''{table}\n''')

    def _generate_stage_and_session_tables(self):
        
        session_row = OrderedDict()

        for stage_summary in self.execution_results.values():

            stage_summary_dict = stage_summary.stage_metrics.dict()

            for field_name in self.session_metrics:
                if session_row.get(field_name) is None:
                    session_row[field_name] = stage_summary_dict.get(field_name)

                else:
                    session_row[field_name] += stage_summary_dict.get(field_name)

            table_row = OrderedDict()
            for field_name in stage_summary.stage_table_header_keys:
                header_name = stage_summary.stage_table_headers.get(field_name)
                table_row[header_name] = stage_summary_dict.get(field_name)

            self.stage_summary_tables['stage_results'].append(table_row)

            if stage_summary.stage_streamed_analytics:
                self._has_streamed = True

                time_steps = []
                current_batch_time = 0

                for time_step in stage_summary.stage_metrics.streamed_batch_timings:
                    current_batch_time += time_step
                    time_steps.append(current_batch_time)
                    
                self._graph_time_steps = time_steps

                self.stage_streamed_data['completed'][stage_summary.stage_metrics.name] = stage_summary.stage_metrics.streamed_completed
                self.stage_streamed_data['succeeded'][stage_summary.stage_metrics.name] = stage_summary.stage_metrics.streamed_succeeded
                self.stage_streamed_data['failed'][stage_summary.stage_metrics.name] = stage_summary.stage_metrics.streamed_failed
                self.stage_streamed_data['completion_rates'][stage_summary.stage_metrics.name] = stage_summary.stage_metrics.streamed_completion_rates

        self.session_table_rows.append(session_row)

    def _generate_actions_and_tasks_table(self):

        for stage_name, stage_metrics in self.execution_results.items():
            for action_or_task_name, group_metrics_set in stage_metrics.action_and_task_metrics.items():

                table_row = OrderedDict()
                table_row['name'] = action_or_task_name
                table_row['stage'] = stage_name

                common_metrics = stage_metrics.common_metrics.get(action_or_task_name)
                common_metrics_data = common_metrics.dict()
                for field_name, field_value in common_metrics_data.items():
                    table_row[field_name] = field_value

                self.actions_and_tasks_common_table_rows.append(table_row)

                for group in stage_metrics.groups:
                    table_row = OrderedDict()
                    table_row['name'] = action_or_task_name
                    table_row['stage'] = stage_name
                
                    metrics = group_metrics_set.get_group(group)
            
                    for field_name, field_value in metrics.items():
                        
                        header_name = field_name.replace(
                            f'{group}_', ''
                        ).replace(
                            '_', ' '
                        )

                        table_row[header_name] = field_value
                    
                    self.actions_and_tasks_table_rows[group].append(table_row)



import uuid
from datetime import datetime
from hedra.core.engines.types.common.types import RequestTypes


try:
    from cassandra.cqlengine import columns
    from cassandra.cqlengine.models import Model
    has_connector = True

except Exception:
    columns = object
    Model = None
    has_connector = False


class CassandraHTTPResultSchema:

    def __init__(self, table_name: str) -> None:
        self.results_columns = {
            'id': columns.UUID(primary_key=True, default=uuid.uuid4),
            'name': columns.Text(min_length=1, index=True),
            'method': columns.Text(min_length=1),
            'url': columns.Text(min_length=1),
            'headers': columns.Text(),
            'data': columns.Text(),
            'error': columns.Text(),
            'status': columns.Integer(),
            'reason': columns.Text(),
            'params': columns.Map(
                columns.Text(),
                columns.Text()
            ),
            'wait_start': columns.Float(),
            'start': columns.Float(),
            'connect_end': columns.Float(),
            'write_end': columns.Float(),
            'complete': columns.Float(),
            'checks': columns.List(
                columns.Text()
            ),
            'user': columns.Text(),
            'tags': columns.List(
                columns.Map(
                    columns.Text(),
                    columns.Text()
                )
            ),
            'created_at': columns.DateTime(default=datetime.now)
        }

        self.results_table = type(
            f'{table_name}_http',
            (Model, ), 
            self.results_columns
        )

        self.type = RequestTypes.HTTP



from typing import Dict, Any, Union
from .http2_processed_result import HTTP2ProcessedResult
from hedra.core.engines.types.graphql_http2 import GraphQLHTTP2Result


class GraphQLHTTP2ProcessedResult(HTTP2ProcessedResult):

    __slots__ = (
        'event_id',
        'action_id',
        'url',
        'ip_addr',
        'method',
        'path',
        'params',
        'hostname',
        'status',
        'headers',
        'data',
        'status',
        'timings',
        'query'
    )

    def __init__(
        self, 
        stage: str, 
        result: GraphQLHTTP2Result
    ) -> None:
        super(GraphQLHTTP2ProcessedResult, self).__init__(
            stage,
            result
        )

        self.query = result.query

    def to_dict(self) -> Dict[str, Union[str, int, float]]:
        graphql_result_dict = super().to_dict()

        return {
            **graphql_result_dict,
            'query': self.query
        }


import asyncio
import random
from hedra.core.engines.types.common.protocols import PingConnection
from hedra.core.engines.types.common.protocols.ping.ping_type import (
    PingTypesMap
)
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.experiments.mutations.types.base.mutation import Mutation
from hedra.core.experiments.mutations.types.base.mutation_type import MutationType
from typing import Union, Tuple
from .validator import InjectPingValidator


Request = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action
]


class InjectPing(Mutation):

    def __init__(
        self, 
        name: str, 
        chance: float,
        *targets: Tuple[str, ...],
        ping_type: str='icmp',
        timeout: int=2
    ) -> None:
        super().__init__(
            name, 
            chance,
            MutationType.INJECT_PING,
            *targets
        )

        validated_mutation = InjectPingValidator(
            ping_type=ping_type,
            timeout=timeout
        )

        self.ping_connection = PingConnection()
        self.types_map = PingTypesMap()
        self.ping_type = self.types_map.get(validated_mutation.ping_type)
        self.timeout = validated_mutation.timeout

    async def mutate(self, action: Request=None) -> Request:
        chance_roll = random.uniform(0, 1)
        if chance_roll <= self.chance:
            return action
        
        try:
            await asyncio.wait_for(
                self.ping_connection.ping(
                    action.url.socket_config,
                    ping_type=self.ping_type,
                    timeout=self.timeout
                ),
                timeout=self.timeout
            )
        except Exception:
            pass

        return action
    
    def copy(self):
        return InjectPing(
            self.name,
            self.chance,
            *list(self.targets),
            ping_type=self.types_map.get_name(
                self.ping_type
            ),
            timeout=self.timeout
        )


import asyncio
import time
import traceback
import uuid
from typing import (
    Dict, 
    Any, 
    Union, 
    Coroutine, 
    TypeVar, 
    Optional
)
from hedra.core.engines.types.common.base_engine import BaseEngine
from hedra.core.engines.types.common.ssl import get_default_ssl_context
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.concurrency import Semaphore
from hedra.core.engines.types.tracing.trace_session import (
    TraceSession, 
    Trace
)

from hedra.logging import HedraLogger
from .connection import HTTPConnection
from .action import HTTPAction
from .result import HTTPResult
from .pool import Pool


A = TypeVar('A')
R = TypeVar('R')


class MercuryHTTPClient(BaseEngine[Union[A, HTTPAction], Union[R, HTTPResult]]):

    __slots__ = (
        'session_id',
        'timeouts',
        'registered',
        '_hosts',
        'closed',
        'sem',
        'pool',
        'active',
        'waiter',
        'ssl_context',
        'logger',
        'tracing_session'
    )

    def __init__(
        self,
        concurrency: int=10**3, 
        timeouts: Timeouts = Timeouts(), 
        reset_connections: bool=False,
        tracing_session: Optional[TraceSession]=None
    ) -> None:
        super(
            MercuryHTTPClient,
            self
        ).__init__()

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self.registered: Dict[str, HTTPAction] = {}
        self._hosts = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=reset_connections)
        self.tracing_session: Union[TraceSession, None] = tracing_session
        self.logger = HedraLogger()
        self.logger.initialize()
        self.pool.create_pool()

        self.active = 0
        self.waiter = None

        self.ssl_context = get_default_ssl_context()

    def config_to_dict(self):
        return {
            'concurrency': self.pool.size,
            'timeouts': {
                'connect_timeout': self.timeouts.connect_timeout,
                'socket_read_timeout': self.timeouts.socket_read_timeout,
                'total_timeout': self.timeouts.total_timeout
            },
            'reset_connections': self.pool.reset_connections
        }

    async def set_pool(self, concurrency: int):
        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=self.pool.reset_connections)
        self.pool.create_pool()

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            self.pool.connections.append(
                HTTPConnection(self.pool.reset_connections)
            )
        
        self.sem = Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.connections = self.pool.connections[:self.pool.size]
        self.sem = Semaphore(self.pool.size)
    
    async def prepare(self, action: HTTPAction) -> Coroutine[Any, Any, None]:
        try:
            if action.url.is_ssl:
                action.ssl_context = self.ssl_context
                
            if self._hosts.get(action.url.hostname) is None:
                socket_configs = await asyncio.wait_for(action.url.lookup(), timeout=self.timeouts.connect_timeout)
            
                for ip_addr, configs in socket_configs.items():
                    for config in configs:

                        connection = HTTPConnection()
                        
                        try:
                            await connection.make_connection(
                                action.url.hostname,
                                ip_addr,
                                action.url.port,
                                config,
                                ssl=action.ssl_context,
                                timeout=self.timeouts.connect_timeout
                            )

                            action.url.socket_config = config
                            action.url.ip_addr = ip_addr
                            action.url.has_ip_addr = True
                            break

                        except Exception as e:
                            pass

                    if action.url.socket_config:
                        break

                if action.url.socket_config is None:
                    raise Exception('Err. - No socket found.')

                self._hosts[action.url.hostname] = {
                    'ip_addr': action.url.ip_addr,
                    'socket_config': action.url.socket_config
                }

            else:
                host_config = self._hosts[action.url.hostname]
                action.url.ip_addr = host_config.get('ip_addr')
                action.url.socket_config = host_config.get('socket_config')

            if action.is_setup is False:
                action.setup()

            self.registered[action.name] = action

        except Exception as e:       
            raise e

    async def execute_prepared_request(self, action: HTTPAction) -> Coroutine[Any, Any, HTTPResult]:
        trace: Union[Trace, None] = None
        if self.tracing_session:
            trace = self.tracing_session.create_trace()
            await trace.on_request_start(action)
  
        response = HTTPResult(action)
        response.wait_start = time.monotonic()
        self.active += 1

        if trace and trace.on_connection_queued_start:
            await trace.on_connection_queued_start(
                trace.span,
                action,
                response
            )
 
        async with self.sem:

            if trace and trace.on_connection_queued_end:
                await trace.on_connection_queued_end(
                    trace.span,
                    action,
                    response
                )
            
            try:
                
                connection = self.pool.connections.pop()

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()
                    
                if action.hooks.before:
                    action = await self.execute_before(action)
                    action.setup()

                response.start = time.monotonic()

                if trace and trace.on_connection_create_start:
                    await trace.on_connection_create_start(
                        trace.span,
                        action,
                        response
                    )


                await connection.make_connection(
                    action.url.hostname,
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    timeout=self.timeouts.connect_timeout,
                    ssl=action.ssl_context
                )

                response.connect_end = time.monotonic()

                if trace and trace.on_connection_create_end:
                    await trace.on_connection_create_end(
                        trace.span,
                        action,
                        response
                    )

                connection.write(action.encoded_headers)

                if trace and trace.on_request_headers_sent:
                    await trace.on_request_headers_sent(
                        trace.span,
                        action,
                        response
                    )
                
                if action.encoded_data:
                    if action.is_stream:
                        action.write_chunks(
                            connection,
                            trace
                        )

                    else:
                        connection.write(action.encoded_data)

                response.write_end = time.monotonic()

                if action.encoded_data and trace and trace.on_request_data_sent:
                        await trace.on_request_data_sent(
                            trace.span,
                            action,
                            response
                        )

                response.response_code = await asyncio.wait_for(
                    connection.reader.readline_fast(),
                    timeout=self.timeouts.socket_read_timeout
                )

                headers = await asyncio.wait_for(
                    connection.read_headers(),
                    timeout=self.timeouts.socket_read_timeout
                )

                if trace and trace.on_response_headers_received:
                    await trace.on_response_headers_received(
                        trace.span,
                        action,
                        response
                    )

                status = response.status
            
                if status >= 300 and status < 400:

                    if trace and trace.on_request_redirect:
                        await trace.on_request_redirect(
                            trace.span,
                            action,
                            response
                        )

                    elapsed_time = 0
                    redirect_time_start = time.time()

                    for _ in range(action.redirects):

                        if elapsed_time > self.timeouts.total_timeout:
                            response.status = 408
                            raise Exception('Request timed out while redirecting.')
                        
                        redirect_url = str(headers.get(b'location'))
                        if redirect_url.startswith('http') is False:
                            action.url.path = redirect_url
                            action.encoded_headers = None
                            action.setup()

                        else:
                            await action.url.replace(redirect_url)
                            action.encoded_headers = None
                            action.is_setup = False
                            await self.prepare(action)

                            await connection.make_connection(
                                action.url.hostname,
                                action.url.ip_addr,
                                action.url.port,
                                action.url.socket_config,
                                timeout=self.timeouts.connect_timeout,
                                ssl=action.ssl_context
                            )

                        response.connect_end = time.monotonic()
                            
                        connection.write(action.encoded_headers)
                        
                        if action.encoded_data:
                            if action.is_stream:
                                action.write_chunks(connection)

                            else:
                                connection.write(action.encoded_data)

                        response.write_end = time.monotonic()

                        response.response_code = await asyncio.wait_for(
                            connection.reader.readline_fast(),
                            timeout=self.timeouts.socket_read_timeout
                        )

                        headers = await asyncio.wait_for(
                            connection.read_headers(),
                            timeout=self.timeouts.socket_read_timeout
                        )

                        status = response.status

                        if status >= 200 and status < 300:
                            break

                        elapsed_time = time.time() - redirect_time_start

                content_length = headers.get(b'content-length')
                transfer_encoding = headers.get(b'transfer-encoding')

                # We require Content-Length or Transfer-Encoding headers to read a
                # request body, otherwise it's anyone's guess as to how big the body
                # is, and we ain't playing that game.
                body = bytearray()
                if content_length:
                    body = await asyncio.wait_for(
                        connection.readexactly(int(content_length)),
                        timeout=self.timeouts.socket_read_timeout
                    )

                elif transfer_encoding:
                    
                    all_chunks_read = False

                    while True and not all_chunks_read:

                        chunk_size = int((await connection.readuntil()).rstrip(), 16)
    
                        if not chunk_size:
                            # read last CRLF
                            body.extend(
                                await asyncio.wait_for(
                                    connection.readuntil(),
                                    timeout=self.timeouts.socket_read_timeout
                                )
                            )
                            
                            break
                        
                        chunk = await asyncio.wait_for(
                            connection.readexactly(chunk_size + 2),
                            timeout=self.timeouts.socket_read_timeout
                        )

                        body.extend(
                            chunk[:-2]
                        )

                        if trace and trace.on_response_chunk_received:
                            await trace.on_response_chunk_received(
                                trace.span,
                                action,
                                response
                            )

                    all_chunks_read = True
         
                response.complete = time.monotonic()

                if trace and trace.on_response_data_received:
                    await trace.on_response_data_received(
                        trace.span,
                        action,
                        response
                    )

                response.headers = headers
                response.body = body
                self.pool.connections.append(connection)

                if action.hooks.after:
                    response = await self.execute_after(action, response)
                    action.setup()

                if action.hooks.checks:
                    response = await self.execute_checks(action, response)

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel.call(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()       

            except Exception as e:
                response.complete = time.monotonic()
                response.error = str(e)

                self.pool.connections.append(HTTPConnection(reset_connection=self.pool.reset_connections))

                if trace and trace.on_request_exception:
                    await trace.on_request_exception(response)

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            if trace and trace.on_request_end:
                await trace.on_request_end(response)

            return response

    async def close(self):
        if self.closed is False:
            await self.pool.close()
            self.closed = True



from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.types import RequestTypes
from typing import (
    Dict,
    Any,
    Union,
    Callable
)
from .parser_types import (
    GraphQLActionParser,
    GraphQLResultParser,
    GraphQLHTTP2ActionParser,
    GraphQLHTTP2ResultParser,
    GRPCActionParser,
    GRPCResultParser,
    HTTPActionParser,
    HTTPResultParser,
    HTTP2ActionParser,
    HTTP2ResultParser,
    HTTP3ActionParser,
    HTTP3ResultParser,
    PlaywrightActionParser,
    PlaywrightResultParser,
    UDPActionParser,
    UDPResultParser,
    WebsocketActionParser,
    WebsocketResultParser
)


class Parser:

    def __init__(self) -> None:
        
        self._action_parsers: Dict[
            str,
            Callable[
                [
                    Dict[str, Any]
                ],
                Union[
                    GraphQLActionParser,
                    GraphQLHTTP2ActionParser,
                    GRPCActionParser,
                    HTTPActionParser,
                    HTTP2ActionParser,
                    HTTP3ActionParser,
                    PlaywrightActionParser,
                    UDPActionParser,
                    WebsocketActionParser
                ]
            ]
        ] = {
            'graphql': lambda config, options: GraphQLActionParser(
                config,
                options
            ),
            'graphqlh2': lambda config, options: GraphQLHTTP2ActionParser(
                config,
                options
            ),
            'grpc': lambda config, options: GRPCActionParser(
                config,
                options
            ),
            'http': lambda config, options: HTTPActionParser(
                config,
                options
            ),
            'http2': lambda config, options: HTTP2ActionParser(
                config,
                options
            ),
            'http3': lambda config, options: HTTP3ActionParser(
                config,
                options
            ),
            'playwright': lambda config, options: PlaywrightActionParser(
                config,
                options
            ),
            'udp': lambda config, options: UDPActionParser(
                config,
                options
            ),
            'websocket': lambda config, options: WebsocketActionParser(
                config,
                options
            )
        }

        self._result_parsers: Dict[
            str,
            Callable[
                [
                    Dict[str, Any]
                ],
                Union[
                    GraphQLResultParser,
                    GraphQLHTTP2ResultParser,
                    GRPCResultParser,
                    HTTPResultParser,
                    HTTP2ResultParser,
                    HTTP3ResultParser,
                    PlaywrightResultParser,
                    UDPResultParser,
                    WebsocketResultParser
                ]
            ]
        ] = {
            'graphql': lambda config, options: GraphQLResultParser(
                config,
                options
            ),
            'graphqlh2': lambda config, options: GraphQLHTTP2ResultParser(
                config,
                options
            ),
            'grpc': lambda config, options: GRPCResultParser(
                config,
                options
            ),
            'http': lambda config, options: HTTPResultParser(
                config,
                options
            ),
            'http2': lambda config, options: HTTP2ResultParser(
                config,
                options
            ),
            'http3': lambda config, options: HTTP3ResultParser(
                config,
                options
            ),
            'playwright': lambda config, options: PlaywrightResultParser(
                config,
                options
            ),
            'udp': lambda config, options: UDPResultParser(
                config,
                options
            ),
            'websocket': lambda config, options: WebsocketResultParser(
                config,
                options
            )
        }

        self._active_action_parsers: Dict[
            str, 
            Union[
                GraphQLActionParser,
                GraphQLHTTP2ActionParser,
                GRPCActionParser,
                HTTPActionParser,
                HTTP2ActionParser,
                HTTP3ActionParser,
                PlaywrightActionParser,
                UDPActionParser,
                WebsocketActionParser
            ]
        ] = {}

        self._active_result_parser: Dict[
            str,
            Union[
                GraphQLResultParser,
                GraphQLHTTP2ResultParser,
                GRPCResultParser,
                HTTPResultParser,
                HTTP2ResultParser,
                HTTP3ResultParser,
                PlaywrightResultParser,
                UDPResultParser,
                WebsocketResultParser
            ]
        ] = {}

    async def parse_action(
        self, 
        action_data: Dict[str, Any],
        stage: str,
        config: Config,
        options: Dict[str, Any]={}
    ):
        engine_type = action_data.get('engine')
        parser = self._active_action_parsers.get(engine_type)

        if parser is None:
            parser = self._action_parsers.get(engine_type)(
                config,
                options
            )

        return await parser.parse(
            action_data,
            stage
        )
    
    async def parse_result(
        self,
        result_data: Dict[str, Any],
        config: Config,
        options: Dict[str, Any]={}
    ):
        
        engine_type = result_data.get('engine')
        parser = self._active_result_parser.get(engine_type)
        if parser is None:
            parser = self._result_parsers.get(engine_type)(
                config,
                options
            )

        return await parser.parse(result_data) 


import click
import psutil
import os
from hedra.cli.graph import (
    check_graph,
    run_graph,
    create_graph
)

@click.group(help='Commands to run, lint, generate, and manage graphs.')
def graph():
    pass


@graph.command(help="Run a specified test file.")
@click.argument('path')
@click.option(
    '--cpus', 
    default=psutil.cpu_count(), 
    help='Number of CPUs to use. Default is the number of physical processesors available to the system.'
)
@click.option(
    '--skip',
    default='',
    help='Comma-delimited list of Stage names to skip.'
)
@click.option(
    '--retries',
    default=0,
    help='Global retries for graph.'
)
@click.option(
    '--show-summaries',
    default='',
    help='Comma-delimited list of results tables to show upon completion.'
)
@click.option(
    '--hide-summaries',
    default='',
    help='Comma-delimited list of results tables to omit upon completion.'
)
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
@click.option(
    '--log-directory',
    default=f'{os.getcwd()}/logs',
    help='Output directory for logfiles. If the directory does not exist it will be created.'
)
@click.option(
    '--bypass-connection-validation',
    is_flag=True,
    show_default=True,
    default=False,
    help="Skip Hedra's action connection validation."
)
@click.option(
    '--connection-validation-retries',
    default=3,
    help="Set the number of retries for connection validation."
)
@click.option(
    '--enable-latest',
    is_flag=True,
    show_default=True,
    default=False,
    help='Enable features marked as unstable.'
)
def run(
    path: str, 
    cpus: int, 
    skip: str,
    retries: int,
    show_summaries: str,
    hide_summaries: str,
    log_level: str, 
    log_directory: str,
    bypass_connection_validation: bool,
    connection_validation_retries: int,
    enable_latest: bool,
):
    run_graph(
        path, 
        cpus,
        skip, 
        retries,
        show_summaries,
        hide_summaries,
        log_level, 
        log_directory,
        bypass_connection_validation,
        connection_validation_retries,
        enable_latest
    )


@graph.command(help="Validate the specified test file.")
@click.argument('path')
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
def check(path: str, log_level: str):
    check_graph(path, log_level)


@graph.command(
    help='Creates basic scaffolding for a test graph at the specified path.'
)
@click.argument('path')
@click.option(
    '--stages',
    help='Optional comma delimited list of stages to generate for the graph.'
)
@click.option(
    '--engine',
    default='http',
    help='Engine to use in generated graph.'
)
@click.option(
    '--reporter',
    default='json',
    help='Reporter to use in generated graph.'
)
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
def create(
    path: str, 
    stages: str, 
    engine: str,
    reporter: str,
    log_level: str,
):
    create_graph(
        path, 
        stages, 
        engine,
        reporter,
        log_level
    )


from __future__ import annotations
import uuid
from typing import (
    List, 
    Dict, 
    Tuple, 
    Any,
    Union
)
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from .analyze.analyze_edge import AnalyzeEdge, BaseEdge
from .common.transtition_metadata import TransitionMetadata
from hedra.core.hooks.types.base.simple_context import SimpleContext
from .common.complete_edge import CompleteEdge
from .common.error_edge import ErrorEdge
from .act.act_edge import ActEdge
from .execute.execute_edge import ExecuteEdge
from .idle.idle_edge import IdleEdge
from .optimize.optimize_edge import OptimizeEdge
from .setup.setup_edge import SetupEdge
from .submit.submit_edge import SubmitEdge


HistoryUpdate = Dict[Tuple[str, str], Any]


class Transition:

    def __init__(self, metadata: TransitionMetadata, from_stage: Stage, to_stage: Stage) -> None:
        self.transition_id = str(uuid.uuid4())
        self.metadata = metadata
        self.from_stage = from_stage
        self.to_stage = to_stage
        self.edges: List[BaseEdge] = []
        self.edges_by_name: Dict[Tuple[str, str], BaseEdge] = {}
        self.adjacency_list: Dict[str, List[Transition]] = []
        self.predecessors = []
        self.descendants = []
        self.destinations: List[str] = []
        self.transition_idx = 0
        self.retries = from_stage.stage_retries
        edge_types = {
            StageTypes.ACT: ActEdge,
            StageTypes.ANALYZE: AnalyzeEdge,
            StageTypes.COMPLETE: CompleteEdge,
            StageTypes.ERROR: ErrorEdge,
            StageTypes.EXECUTE: ExecuteEdge,
            StageTypes.IDLE: IdleEdge,
            StageTypes.OPTIMIZE: OptimizeEdge,
            StageTypes.SETUP: SetupEdge,
            StageTypes.SUBMIT: SubmitEdge,
        }

        self.edge: BaseEdge = edge_types.get(from_stage.stage_type)(
            from_stage,
            to_stage
        )

    async def execute(self):

        self.edge.setup()

        if self.edge.source.context is None:
            self.edge.source.context = SimpleContext()
        
        result: Union[None, Tuple[None, StageTypes]] = None
        if self.retries > 0:
            for _ in range(self.retries):
                result = await self.edge.transition()

                if self.edge.exception is None:
                    break
            
        else:
            result = await self.edge.transition()

        self.edge.descendants = {
            descendant: self.edges_by_name.get((
                self.edge.source.name,
                descendant
            )) for descendant in self.descendants
        }

        skip_next_stages = [
            StageTypes.COMPLETE,
            StageTypes.ERROR,
        ]

        is_ignored_stage = self.to_stage.stage_type in skip_next_stages
        stage_skipped = self.edge.source.skip is True and self.edge.skip_stage is False
        invalid_transition = self.metadata.is_valid is False
        has_exception = self.edge.exception is not None

        pass_to_next = (
            is_ignored_stage or stage_skipped or invalid_transition or has_exception
        ) is False

        if pass_to_next:
            source_name = self.edge.source.name
            selected_edge = self.edge

            if self.edge.skip_stage:
                selected_edge_idx = min([
                    edge.transition_idx for edge in self.edges
                ])

                selected_edge = self.edges[selected_edge_idx]
                source_name = self.edges[selected_edge_idx].source.name

            destination_name = self.edge.destination.name

            neighbors: List[Tuple[str, str]]  = []
            transition_source_histories: Dict[str, HistoryUpdate] = {}

            source_edge_name = (
                source_name, 
                destination_name
            )

            transition_source_history: HistoryUpdate = selected_edge.next_history.get(
                source_edge_name, {}
            )

            for destination in self.destinations:
                for transition in self.adjacency_list[destination]:

                    destination_edge_name = (
                        destination, 
                        transition.edge.destination.name
                    )

                    neighbors.extend([
                        (
                            destination, 
                            transition.edge.destination.name
                        ) for transition in self.adjacency_list[destination]
                    ])

                    transition_source_histories[destination_edge_name] = transition_source_history

            for neighbor in neighbors:
                required_keys = self.edges_by_name[neighbor].requires

                if source_name not in self.edges_by_name[neighbor].from_stage_names:
                    self.edges_by_name[neighbor].from_stage_names.append(source_name)
                    

                for edge_name in self.edge.next_history:
                    source_history: HistoryUpdate = self.edge.next_history[edge_name]

                    self.edges_by_name[neighbor].history.update({
                        edge_name: {
                            key: value for key, value in source_history.items() if key in required_keys
                        }
                    })

                neighbor_edge_source = self.edges_by_name[neighbor].source.name

                previous_edge = (source_name, neighbor_edge_source)

                for history in transition_source_histories.values():
                    if self.edges_by_name[neighbor].history.get(previous_edge) is None and len(history) > 0:
                        self.edges_by_name[neighbor].history.update({
                                previous_edge: {
                                    key: value for key, value in history.items() if key in required_keys
                                }
                            })


        self.edge.edge_data = {}

        return result
